"""
Test formatting of std::unordered_map related structures.
"""

import lldb
from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *
from lldbsuite.test import lldbutil


class StdUnorderedMapDataFormatterTestCase(TestBase):
    SHARED_BUILD_TESTCASE = False

    def check_ptr_or_ref(self, var_name: str):
        var = self.frame().FindVariable(var_name)
        self.assertTrue(var)

        pair = var.GetChildAtIndex(0)
        self.assertTrue(pair)

        self.assertEqual(pair.GetChildAtIndex(0).summary, '"Hello"')
        self.assertEqual(pair.GetChildAtIndex(1).summary, '"World"')

    def check_ptr_ptr(self, var_name: str):
        var = self.frame().FindVariable(var_name)
        self.assertTrue(var)

        ptr = var.GetChildAtIndex(0)
        self.assertTrue(ptr)

        pair = ptr.GetChildAtIndex(0)
        self.assertTrue(pair)

        self.assertEqual(pair.GetChildAtIndex(0).summary, '"Hello"')
        self.assertEqual(pair.GetChildAtIndex(1).summary, '"World"')

    def do_test(self):
        """Test that std::unordered_map related structures are formatted correctly when printed.
        Currently only tests format of std::unordered_map iterators.
        """
        (self.target, process, thread, bkpt) = lldbutil.run_to_source_breakpoint(
            self, "Break here", lldb.SBFileSpec("main.cpp", False)
        )

        # Test empty iterators
        self.expect_expr("empty_iter", "")
        self.expect_expr("const_empty_iter", "")

        lldbutil.continue_to_breakpoint(process, bkpt)

        # Check that key/value is correctly formatted
        self.expect_expr(
            "foo",
            result_children=[
                ValueCheck(name="first", summary='"Foo"'),
                ValueCheck(name="second", summary='"Bar"'),
            ],
        )

        # Check invalid iterator is empty
        self.expect_expr("invalid", "")

        # Const key/val iterator
        self.expect_expr(
            "const_baz",
            result_children=[
                ValueCheck(name="first", summary='"Baz"'),
                ValueCheck(name="second", summary='"Qux"'),
            ],
        )

        # Bucket iterators
        # I.e., std::__hash_map_const_iterator<const_local_iterator<...>>
        # and std::__hash_map_iterator<local_iterator<...>>
        self.expect_expr(
            "bucket_it",
            result_children=[
                ValueCheck(name="first", summary='"Baz"'),
                ValueCheck(name="second", summary='"Qux"'),
            ],
        )

        self.expect_expr(
            "const_bucket_it",
            result_children=[
                ValueCheck(name="first", summary='"Baz"'),
                ValueCheck(name="second", summary='"Qux"'),
            ],
        )

        lldbutil.continue_to_breakpoint(process, bkpt)

        # Test references to std::unordered_map
        self.check_ptr_or_ref("ref1")
        self.check_ptr_or_ref("ref2")
        self.check_ptr_or_ref("ref3")
        self.check_ptr_or_ref("ref4")
        self.check_ptr_or_ref("ref5")
        self.check_ptr_or_ref("ref6")

        # FIXME: we're getting this wrong.
        self.expect_var_path(
            "ref7",
            summary="size=0",
            type="const StringMapT *const &",
        )

    @add_test_categories(["libc++"])
    def test_libcxx(self):
        self.build(dictionary={"USE_LIBCPP": 1})
        self.do_test()

    def do_test_ptr(self):
        """
        Test that pointers to std::unordered_map are formatted correctly.
        """

        (self.target, process, thread, bkpt) = lldbutil.run_to_source_breakpoint(
            self, "Stop here", lldb.SBFileSpec("main.cpp", False)
        )

        self.check_ptr_or_ref("ptr1")
        self.check_ptr_or_ref("ptr2")
        self.check_ptr_or_ref("ptr3")
        self.check_ptr_ptr("ptr4")
        self.check_ptr_ptr("ptr5")
        self.check_ptr_ptr("ptr6")

    @add_test_categories(["libc++"])
    def test_ptr_libcxx(self):
        self.build(dictionary={"USE_LIBCPP": 1})
        self.do_test_ptr()



"""Test Python APIs for target (launch and attach), breakpoint, and process."""


import os

import lldb
from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *
import lldbsuite.test.lldbutil as lldbutil


class HelloWorldTestCase(TestBase):
    SHARED_BUILD_TESTCASE = False
    NO_DEBUG_INFO_TESTCASE = True

    def setUp(self):
        # Call super's setUp().
        TestBase.setUp(self)
        # Find a couple of the line numbers within main.c.
        self.line1 = line_number("main.c", "// Set break point at this line.")
        self.line2 = line_number("main.c", "// Waiting to be attached...")

    def tearDown(self):
        # Destroy process before TestBase.tearDown()
        self.dbg.GetSelectedTarget().GetProcess().Destroy()
        # Call super's tearDown().
        TestBase.tearDown(self)

    @skipIfiOSSimulator
    def test_with_process_launch_api(self):
        """Create target, breakpoint, launch a process, and then kill it."""
        # Get the full path to our executable to be attached/debugged.
        exe = "%s_%d" % (self.getBuildArtifact(self.testMethodName), os.getpid())
        d = {"EXE": exe}
        self.build(dictionary=d)
        self.setTearDownCleanup(dictionary=d)
        target = self.dbg.CreateTarget(exe)

        breakpoint = target.BreakpointCreateByLocation("main.c", self.line1)

        # The default state after breakpoint creation should be enabled.
        self.assertTrue(
            breakpoint.IsEnabled(), "Breakpoint should be enabled after creation"
        )

        breakpoint.SetEnabled(False)
        self.assertTrue(
            not breakpoint.IsEnabled(), "Breakpoint.SetEnabled(False) works"
        )

        breakpoint.SetEnabled(True)
        self.assertTrue(breakpoint.IsEnabled(), "Breakpoint.SetEnabled(True) works")

        # rdar://problem/8364687
        # SBTarget.Launch() issue (or is there some race condition)?

        process = target.LaunchSimple(None, None, self.get_process_working_directory())
        # The following isn't needed anymore, rdar://8364687 is fixed.
        #
        # Apply some dances after LaunchProcess() in order to break at "main".
        # It only works sometimes.
        # self.breakAfterLaunch(process, "main")

        process = target.GetProcess()
        self.assertTrue(process, PROCESS_IS_VALID)

        thread = lldbutil.get_stopped_thread(process, lldb.eStopReasonBreakpoint)
        self.assertIsNotNone(thread)

        # The breakpoint should have a hit count of 1.
        self.assertEqual(breakpoint.GetHitCount(), 1, BREAKPOINT_HIT_ONCE)

    @expectedFailureAll(oslist=["windows"], archs=["aarch64"])
    @skipIfiOSSimulator
    def test_with_attach_to_process_with_id_api(self):
        """Create target, spawn a process, and attach to it with process id."""
        exe = "%s_%d" % (self.testMethodName, os.getpid())
        d = {"EXE": exe}
        self.build(dictionary=d)
        self.setTearDownCleanup(dictionary=d)
        target = self.dbg.CreateTarget(self.getBuildArtifact(exe))

        # Spawn a new process
        token = exe + ".token"
        if not lldb.remote_platform:
            token = self.getBuildArtifact(token)
            if os.path.exists(token):
                os.remove(token)
        popen = self.spawnSubprocess(self.getBuildArtifact(exe), [token])
        lldbutil.wait_for_file_on_target(self, token)

        listener = lldb.SBListener("my.attach.listener")
        error = lldb.SBError()
        process = target.AttachToProcessWithID(listener, popen.pid, error)

        self.assertTrue(error.Success() and process, PROCESS_IS_VALID)

        # Let's check the stack traces of the attached process.
        stacktraces = lldbutil.print_stacktraces(process, string_buffer=True)
        self.expect(
            stacktraces, exe=False, substrs=["main.c:%d" % self.line2, "(int)argc=2"]
        )

    @expectedFailureAll(oslist=["windows"], archs=["aarch64"])
    @skipIfiOSSimulator
    @skipIfAsan  # FIXME: Hangs indefinitely.
    def test_with_attach_to_process_with_name_api(self):
        """Create target, spawn a process, and attach to it with process name."""
        exe = "%s_%d" % (self.testMethodName, os.getpid())
        d = {"EXE": exe}
        self.build(dictionary=d)
        self.setTearDownCleanup(dictionary=d)
        target = self.dbg.CreateTarget(self.getBuildArtifact(exe))

        # Spawn a new process.
        token = exe + ".token"
        if not lldb.remote_platform:
            token = self.getBuildArtifact(token)
            if os.path.exists(token):
                os.remove(token)
        popen = self.spawnSubprocess(self.getBuildArtifact(exe), [token])
        lldbutil.wait_for_file_on_target(self, token)

        listener = lldb.SBListener("my.attach.listener")
        error = lldb.SBError()
        # Pass 'False' since we don't want to wait for new instance of
        # "hello_world" to be launched.
        name = os.path.basename(exe)

        # While we're at it, make sure that passing a None as the process name
        # does not hang LLDB.
        target.AttachToProcessWithName(listener, None, False, error)
        # Also boundary condition test ConnectRemote(), too.
        target.ConnectRemote(listener, None, None, error)

        process = target.AttachToProcessWithName(listener, name, False, error)
        self.assertSuccess(error)
        self.assertTrue(process, PROCESS_IS_VALID)

        # Verify that after attach, our selected target indeed matches name.
        self.expect(
            self.dbg.GetSelectedTarget().GetExecutable().GetFilename(),
            exe=False,
            startstr=name,
        )

        # Let's check the stack traces of the attached process.
        stacktraces = lldbutil.print_stacktraces(process, string_buffer=True)
        self.expect(
            stacktraces, exe=False, substrs=["main.c:%d" % self.line2, "(int)argc=2"]
        )



import lldb
import time
from lldbsuite.test.lldbtest import *
from lldbsuite.test.decorators import *
from lldbsuite.test.gdbclientutils import *
from lldbsuite.test.lldbgdbclient import GDBRemoteTestBase


class TestPlatformKill(GDBRemoteTestBase):
    SHARED_BUILD_TESTCASE = False

    @skipIfRemote
    def test_kill_different_platform(self):
        """Test connecting to a remote linux platform"""

        self.build(dictionary={"CXX_SOURCES": "sleep.cpp"})
        host_process = self.spawnSubprocess(self.getBuildArtifact())

        # Create a fake remote process with the same PID as host_process
        class MyResponder(MockGDBServerResponder):
            def __init__(self):
                MockGDBServerResponder.__init__(self)
                self.got_kill = False

            def qC(self):
                return "QC%x" % host_process.pid

            def k(self):
                self.got_kill = True
                return "X09"

        self.server.responder = MyResponder()

        error = lldb.SBError()
        target = self.dbg.CreateTarget(
            "", "x86_64-pc-linux", "remote-linux", False, error
        )
        self.assertSuccess(error)
        process = self.connect(target)
        self.assertEqual(process.GetProcessID(), host_process.pid)

        host_platform = lldb.SBPlatform("host")
        self.assertSuccess(host_platform.Kill(host_process.pid))

        # Host dies, remote process lives.
        self.assertFalse(self.server.responder.got_kill)
        self.assertIsNotNone(host_process.wait(timeout=10))

        # Now kill the remote one as well
        self.assertSuccess(process.Kill())
        self.assertTrue(self.server.responder.got_kill)



"""Test that the 'add-dsym', aka 'target symbols add', succeeds in the middle of debug session."""


import lldb
from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *
from lldbsuite.test import lldbutil


@skipUnlessDarwin
class AddDsymMidExecutionCommandCase(TestBase):
    SHARED_BUILD_TESTCASE = False

    def setUp(self):
        # Call super's setUp().
        TestBase.setUp(self)
        self.source = "main.c"

    @no_debug_info_test  # Prevent the genaration of the dwarf version of this test
    def test_add_dsym_mid_execution(self):
        """Test that add-dsym mid-execution loads the symbols at the right place for a slid binary."""
        self.build(debug_info="dsym")
        exe = self.getBuildArtifact("a.out")

        self.target = self.dbg.CreateTarget(exe)
        self.assertTrue(self.target, VALID_TARGET)

        main_bp = self.target.BreakpointCreateByName("main", "a.out")
        self.assertTrue(main_bp, VALID_BREAKPOINT)

        self.runCmd("settings set target.disable-aslr false")
        self.process = self.target.LaunchSimple(
            None, None, self.get_process_working_directory()
        )
        self.assertTrue(self.process, PROCESS_IS_VALID)

        # The stop reason of the thread should be breakpoint.
        self.assertState(
            self.process.GetState(), lldb.eStateStopped, STOPPED_DUE_TO_BREAKPOINT
        )

        self.runCmd("add-dsym " + self.getBuildArtifact("hide.app/Contents/a.out.dSYM"))

        self.expect("frame select", substrs=["a.out`main at main.c"])



import lldb


class stop_handler:
    def __init__(self, target, extra_args, dict):
        self.extra_args = extra_args
        self.target = target
        self.counter = 0
        ret_val = self.extra_args.GetValueForKey("return_false")
        if ret_val:
            self.ret_val = False
        else:
            self.ret_val = True

    def handle_stop(self, exe_ctx, stream):
        self.counter += 1
        stream.Print("I have stopped %d times.\n" % (self.counter))
        increment = 1
        value = self.extra_args.GetValueForKey("increment")
        if value:
            increment = value.GetUnsignedIntegerValue()
        else:
            stream.Print("Could not find increment in extra_args\n")
        frame = exe_ctx.GetFrame()
        expression = "g_var += %d" % (increment)
        expr_result = frame.EvaluateExpression(expression)
        if not expr_result.GetError().Success():
            stream.Print(
                "Error running expression: %s" % (expr_result.GetError().GetCString())
            )
        value = exe_ctx.target.FindFirstGlobalVariable("g_var")
        if not value.IsValid():
            stream.Print("Didn't get a valid value for g_var.")
        else:
            int_val = value.GetValueAsUnsigned()
        stream.Print("Returning value: %d from handle_stop.\n" % (self.ret_val))
        return self.ret_val


class bad_handle_stop:
    def __init__(self, target, extra_args, dict):
        print("I am okay")

    def handle_stop(self):
        print("I am bad")


class no_handle_stop:
    def __init__(self, target, extra_args, dict):
        print("I am okay")


class self_deleting_stop:
    def __init__(self, target, extra_args, dict):
        self.target = target

    def handle_stop(self, exe_ctx, stream):
        interp = exe_ctx.target.debugger.GetCommandInterpreter()
        result = lldb.SBCommandReturnObject()
        interp.HandleCommand("handle_id", result)
        id_str = result.GetOutput().rstrip()

        command = f"target stop-hook delete {id_str}"
        interp.HandleCommand(command, result)


stop_hook_id = 0


def handle_stop_hook_id(debugger, command, exe_ctx, result, extra_args):
    global stop_hook_id
    if command == "":
        result.AppendMessage(str(stop_hook_id))
    else:
        stop_hook_id = int(command)



# RUN: %PYTHON -m mlir.dialects.linalg.opdsl.dump_oplib --file %s | FileCheck %s

from mlir.dialects.linalg.opdsl.lang import *


# CHECK: ---
# CHECK-LABEL: matmul
# CHECK: args:
# CHECK:     name: A
# CHECK:     kind: input_tensor
# CHECK:     type_var: T
# CHECK:     shape_map: affine_map<()[s0, s1, s2] -> (s0, s1)>
# CHECK:     name: B
# CHECK:     kind: input_tensor
# CHECK:     type_var: T
# CHECK:     shape_map: affine_map<()[s0, s1, s2] -> (s1, s2)>
# CHECK:     name: C
# CHECK:     kind: output_tensor
# CHECK:     type_var: U
# CHECK:     shape_map: affine_map<()[s0, s1, s2] -> (s0, s2)>
# CHECK:     name: bfn
# CHECK:     kind: binary_fn_attr
# CHECK:     default_fn: mul
# CHECK:     name: ufn
# CHECK:     kind: unary_fn_attr
# CHECK:     default_fn: exp
# CHECK:     name: cast
# CHECK:     kind: type_fn_attr
# CHECK:     default_fn: cast_signed
@linalg_structured_op
def matmul(
    A=TensorDef(T, S.M, S.K),
    B=TensorDef(T, S.K, S.N),
    C=TensorDef(U, S.M, S.N, output=True),
    bfn=BinaryFnAttrDef(default=BinaryFn.mul),
    ufn=UnaryFnAttrDef(default=UnaryFn.exp),
    cast=TypeFnAttrDef(default=TypeFn.cast_signed),
):
    C[D.m, D.n] += bfn(cast(U, A[D.m, D.k]), cast(U, B[D.k, D.n]))


# CHECK: ---
# CHECK-LABEL: fill
# CHECK: args:
# CHECK:     name: value
# CHECK:     kind: scalar
# CHECK-NOT: shape_map:
# CHECK:     type_var: T
@linalg_structured_op
def fill(value=ScalarDef(T), O=TensorDef(T, S.M, S.K, output=True)):
    O[D.m, D.n] = value


# CHECK: ---
# CHECK-LABEL: strided_copy
# CHECK: args:
# CHECK:     name: I
# CHECK:     kind: input_tensor
# CHECK:     type_var: T
# CHECK:     shape_map: affine_map<()[s0, s1, s2, s3, s4, s5] -> (s0, s1)>
# CHECK:     name: O
# CHECK:     kind: output_tensor
# CHECK:     type_var: T
# CHECK:     shape_map: affine_map<()[s0, s1, s2, s3, s4, s5] -> (s2, s3)>
# CHECK:     name: strides
# CHECK:     kind: index_attr
# CHECK:     index_attr_map: affine_map<()[s0, s1, s2, s3, s4, s5] -> (s4, s5)>
# CHECK:     default_indices:
# CHECK:     - 1
# CHECK:     - 2
@linalg_structured_op
def strided_copy(
    I=TensorDef(T, S.IH, S.IW),
    O=TensorDef(T, S.OH, S.OW, output=True),
    strides=IndexAttrDef(S.SH, S.SW, default=[1, 2]),
):
    O[D.oh, D.ow] = I[D.oh * S.SH, D.ow * S.SW]



# RUN: %PYTHON %s | FileCheck %s
from unittest import result
from mlir.ir import (
    Context,
    FunctionType,
    Location,
    Module,
    InsertionPoint,
    IntegerType,
    IndexType,
    MemRefType,
    F32Type,
    Block,
    ArrayAttr,
    Attribute,
    UnitAttr,
    StringAttr,
    DenseI32ArrayAttr,
    ShapedType,
)
from mlir.dialects import openacc, func, arith, memref
from mlir.extras import types


def run(f):
    print("\n// TEST:", f.__name__)
    with Context(), Location.unknown():
        f()
    return f


@run
def testParallelMemcpy():
    module = Module.create()

    dynamic = ShapedType.get_dynamic_size()
    memref_f32_1d_any = MemRefType.get([dynamic], types.f32())

    with InsertionPoint(module.body):
        function_type = FunctionType.get(
            [memref_f32_1d_any, memref_f32_1d_any, types.i64()], []
        )
        f = func.FuncOp(
            type=function_type,
            name="memcpy_idiom",
        )
        f.attributes["sym_visibility"] = StringAttr.get("public")

    with InsertionPoint(f.add_entry_block()):
        c1024 = arith.ConstantOp(types.i32(), 1024)
        c128 = arith.ConstantOp(types.i32(), 128)

        arg0, arg1, arg2 = f.arguments

        copied = openacc.copyin(
            acc_var=arg0.type,
            var=arg0,
            var_type=types.f32(),
            bounds=[],
            async_operands=[],
            implicit=False,
            structured=True,
        )
        created = openacc.create_(
            acc_var=arg1.type,
            var=arg1,
            var_type=types.f32(),
            bounds=[],
            async_operands=[],
            implicit=False,
            structured=True,
        )

        parallel_op = openacc.ParallelOp(
            asyncOperands=[],
            waitOperands=[],
            numGangs=[c1024],
            numWorkers=[],
            vectorLength=[c128],
            reductionOperands=[],
            privateOperands=[],
            firstprivateOperands=[],
            dataClauseOperands=[],
        )

        # Set required device_type and segment attributes to satisfy verifier
        acc_device_none = ArrayAttr.get([Attribute.parse("#acc.device_type<none>")])
        parallel_op.numGangsDeviceType = acc_device_none
        parallel_op.numGangsSegments = DenseI32ArrayAttr.get([1])
        parallel_op.vectorLengthDeviceType = acc_device_none

        parallel_block = Block.create_at_start(parent=parallel_op.region, arg_types=[])

        with InsertionPoint(parallel_block):
            c0 = arith.ConstantOp(types.i64(), 0)
            c1 = arith.ConstantOp(types.i64(), 1)

            loop_op = openacc.LoopOp(
                results_=[],
                lowerbound=[c0],
                upperbound=[f.arguments[2]],
                step=[c1],
                gangOperands=[],
                workerNumOperands=[],
                vectorOperands=[],
                tileOperands=[],
                cacheOperands=[],
                privateOperands=[],
                reductionOperands=[],
                firstprivateOperands=[],
            )

            # Set loop attributes: gang and independent on device_type<none>
            acc_device_none = ArrayAttr.get([Attribute.parse("#acc.device_type<none>")])
            loop_op.gang = acc_device_none
            loop_op.independent = acc_device_none

            loop_block = Block.create_at_start(
                parent=loop_op.region, arg_types=[types.i64()]
            )

            with InsertionPoint(loop_block):
                idx = arith.index_cast(out=IndexType.get(), in_=loop_block.arguments[0])
                val = memref.load(memref=copied, indices=[idx])
                memref.store(value=val, memref=created, indices=[idx])
                openacc.YieldOp([])

            openacc.YieldOp([])

        deleted = openacc.delete(
            acc_var=copied,
            bounds=[],
            async_operands=[],
            implicit=False,
            structured=True,
        )
        copied = openacc.copyout(
            acc_var=created,
            var=arg1,
            var_type=types.f32(),
            bounds=[],
            async_operands=[],
            implicit=False,
            structured=True,
        )
        func.ReturnOp([])

    print(module)

    # CHECK: TEST: testParallelMemcpy
    # CHECK-LABEL:   func.func public @memcpy_idiom(
    # CHECK-SAME:      %[[ARG0:.*]]: memref<?xf32>, %[[ARG1:.*]]: memref<?xf32>, %[[ARG2:.*]]: i64) {
    # CHECK:           %[[CONSTANT_0:.*]] = arith.constant 1024 : i32
    # CHECK:           %[[CONSTANT_1:.*]] = arith.constant 128 : i32
    # CHECK:           %[[COPYIN_0:.*]] = acc.copyin varPtr(%[[ARG0]] : memref<?xf32>) -> memref<?xf32>
    # CHECK:           %[[CREATE_0:.*]] = acc.create varPtr(%[[ARG1]] : memref<?xf32>) -> memref<?xf32>
    # CHECK:           acc.parallel num_gangs({%[[CONSTANT_0]] : i32}) vector_length(%[[CONSTANT_1]] : i32) {
    # CHECK:             %[[CONSTANT_2:.*]] = arith.constant 0 : i64
    # CHECK:             %[[CONSTANT_3:.*]] = arith.constant 1 : i64
    # CHECK:             acc.loop gang control(%[[VAL_0:.*]] : i64) = (%[[CONSTANT_2]] : i64) to (%[[ARG2]] : i64)  step (%[[CONSTANT_3]] : i64) {
    # CHECK:               %[[INDEX_CAST_0:.*]] = arith.index_cast %[[VAL_0]] : i64 to index
    # CHECK:               %[[LOAD_0:.*]] = memref.load %[[COPYIN_0]]{{\[}}%[[INDEX_CAST_0]]] : memref<?xf32>
    # CHECK:               memref.store %[[LOAD_0]], %[[CREATE_0]]{{\[}}%[[INDEX_CAST_0]]] : memref<?xf32>
    # CHECK:               acc.yield
    # CHECK:             } attributes {independent = [#acc.device_type<none>]}
    # CHECK:             acc.yield
    # CHECK:           }
    # CHECK:           acc.delete accPtr(%[[COPYIN_0]] : memref<?xf32>)
    # CHECK:           acc.copyout accPtr(%[[CREATE_0]] : memref<?xf32>) to varPtr(%[[ARG1]] : memref<?xf32>)
    # CHECK:           return
    # CHECK:         }



"""
LLDB AppKit formatters

Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
See https://llvm.org/LICENSE.txt for license information.
SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
"""

# summary provider for CF(Mutable)BitVector
import lldb
import ctypes
import lldb.runtime.objc.objc_runtime
import lldb.formatters.metrics
import lldb.formatters.Logger

# first define some utility functions


def byte_index(abs_pos):
    logger = lldb.formatters.Logger.Logger()
    return abs_pos / 8


def bit_index(abs_pos):
    logger = lldb.formatters.Logger.Logger()
    return abs_pos & 7


def get_bit(byte, index):
    logger = lldb.formatters.Logger.Logger()
    if index < 0 or index > 7:
        return None
    return (byte >> (7 - index)) & 1


def grab_array_item_data(pointer, index):
    logger = lldb.formatters.Logger.Logger()
    return pointer.GetPointeeData(index, 1)


statistics = lldb.formatters.metrics.Metrics()
statistics.add_metric("invalid_isa")
statistics.add_metric("invalid_pointer")
statistics.add_metric("unknown_class")
statistics.add_metric("code_notrun")

# despite the similary to synthetic children providers, these classes are not
# trying to provide anything but a summary for a CF*BitVector, so they need not
# obey the interface specification for synthetic children providers


class CFBitVectorKnown_SummaryProvider:
    def adjust_for_architecture(self):
        logger = lldb.formatters.Logger.Logger()
        self.uiint_size = self.sys_params.types_cache.NSUInteger.GetByteSize()
        pass

    def __init__(self, valobj, params):
        logger = lldb.formatters.Logger.Logger()
        self.valobj = valobj
        self.sys_params = params
        if not (self.sys_params.types_cache.NSUInteger):
            if self.sys_params.is_64_bit:
                self.sys_params.types_cache.NSUInteger = (
                    self.valobj.GetType().GetBasicType(lldb.eBasicTypeUnsignedLong)
                )
            else:
                self.sys_params.types_cache.NSUInteger = (
                    self.valobj.GetType().GetBasicType(lldb.eBasicTypeUnsignedInt)
                )
        if not (self.sys_params.types_cache.charptr):
            self.sys_params.types_cache.charptr = (
                self.valobj.GetType().GetBasicType(lldb.eBasicTypeChar).GetPointerType()
            )
        self.update()

    def update(self):
        logger = lldb.formatters.Logger.Logger()
        self.adjust_for_architecture()

    # we skip the CFRuntimeBase
    # then the next CFIndex is the count
    # then we skip another CFIndex and then we get at a byte array
    # that wraps the individual bits

    def contents(self):
        logger = lldb.formatters.Logger.Logger()
        count_vo = self.valobj.CreateChildAtOffset(
            "count",
            self.sys_params.cfruntime_size,
            self.sys_params.types_cache.NSUInteger,
        )
        count = count_vo.GetValueAsUnsigned(0)
        if count == 0:
            return "(empty)"

        array_vo = self.valobj.CreateChildAtOffset(
            "data",
            self.sys_params.cfruntime_size + 2 * self.uiint_size,
            self.sys_params.types_cache.charptr,
        )

        data_list = []
        cur_byte_pos = None
        for i in range(0, count):
            if cur_byte_pos is None:
                cur_byte_pos = byte_index(i)
                cur_byte = grab_array_item_data(array_vo, cur_byte_pos)
                cur_byte_val = cur_byte.uint8[0]
            else:
                byte_pos = byte_index(i)
                # do not fetch the pointee data every single time through
                if byte_pos != cur_byte_pos:
                    cur_byte_pos = byte_pos
                    cur_byte = grab_array_item_data(array_vo, cur_byte_pos)
                    cur_byte_val = cur_byte.uint8[0]
            bit = get_bit(cur_byte_val, bit_index(i))
            if (i % 4) == 0:
                data_list.append(" ")
            if bit == 1:
                data_list.append("1")
            else:
                data_list.append("0")
        return "".join(data_list)


class CFBitVectorUnknown_SummaryProvider:
    def adjust_for_architecture(self):
        pass

    def __init__(self, valobj, params):
        logger = lldb.formatters.Logger.Logger()
        self.valobj = valobj
        self.sys_params = params
        self.update()

    def update(self):
        logger = lldb.formatters.Logger.Logger()
        self.adjust_for_architecture()

    def contents(self):
        logger = lldb.formatters.Logger.Logger()
        return "<unable to summarize this CFBitVector>"


def GetSummary_Impl(valobj):
    logger = lldb.formatters.Logger.Logger()
    global statistics
    (
        class_data,
        wrapper,
    ) = lldb.runtime.objc.objc_runtime.Utilities.prepare_class_detection(
        valobj, statistics
    )
    if wrapper:
        return wrapper

    name_string = class_data.class_name()
    actual_name = name_string

    logger >> "name string got was " + str(name_string) + " but actual name is " + str(
        actual_name
    )

    if class_data.is_cftype():
        # CFBitVectorRef does not expose an actual NSWrapper type, so we have to check that this is
        # an NSCFType and then check we are a pointer-to CFBitVectorRef
        valobj_type = valobj.GetType()
        if valobj_type.IsValid() and valobj_type.IsPointerType():
            valobj_type = valobj_type.GetPointeeType()
            if valobj_type.IsValid():
                actual_name = valobj_type.GetName()
        if actual_name == "__CFBitVector" or actual_name == "__CFMutableBitVector":
            wrapper = CFBitVectorKnown_SummaryProvider(valobj, class_data.sys_params)
            statistics.metric_hit("code_notrun", valobj)
        else:
            wrapper = CFBitVectorUnknown_SummaryProvider(valobj, class_data.sys_params)
            print(actual_name)
    else:
        wrapper = CFBitVectorUnknown_SummaryProvider(valobj, class_data.sys_params)
        print(name_string)
        statistics.metric_hit(
            "unknown_class", valobj.GetName() + " seen as " + name_string
        )
    return wrapper


def CFBitVector_SummaryProvider(valobj, dict):
    logger = lldb.formatters.Logger.Logger()
    provider = GetSummary_Impl(valobj)
    if provider is not None:
        if isinstance(
            provider, lldb.runtime.objc.objc_runtime.SpecialSituation_Description
        ):
            return provider.message()
        try:
            summary = provider.contents()
        except:
            summary = None
        logger >> "summary got from provider: " + str(summary)
        if summary is None or summary == "":
            summary = "<variable is not CFBitVector>"
        return summary
    return "Summary Unavailable"


def __lldb_init_module(debugger, dict):
    debugger.HandleCommand(
        "type summary add -F CFBitVector.CFBitVector_SummaryProvider CFBitVectorRef CFMutableBitVectorRef"
    )



"""
Test some lldb help commands.

See also CommandInterpreter::OutputFormattedHelpText().
"""


import os
import lldb
from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *
from lldbsuite.test import lldbutil


class HelpCommandTestCase(TestBase):
    @no_debug_info_test
    def test_simplehelp(self):
        """A simple test of 'help' command and its output."""
        self.expect("help", startstr="Debugger commands:")

        self.expect("help -a", matching=False, substrs=["next"])

        self.expect("help", matching=True, substrs=["next"])

    @no_debug_info_test
    def test_help_on_help(self):
        """Testing the help on the help facility."""
        self.expect(
            "help help",
            matching=True,
            substrs=["--hide-aliases", "--hide-user-commands"],
        )

    @no_debug_info_test
    def test_help_arch(self):
        """Test 'help arch' which should list of supported architectures."""
        self.expect("help arch", substrs=["arm", "i386", "x86_64"])

    @no_debug_info_test
    def test_help_version(self):
        """Test 'help version' and 'version' commands."""
        self.expect("help version", substrs=["Show the LLDB debugger version."])
        self.expect("version", patterns=["lldb( version|-[0-9]+).*\n"])

    @no_debug_info_test
    def test_help_should_not_crash_lldb(self):
        """Command 'help disasm' should not crash lldb."""
        self.runCmd("help disasm", check=False)
        self.runCmd("help unsigned-integer")

    @no_debug_info_test
    def test_help_memory_read_should_not_crash_lldb(self):
        """Command 'help memory read' should not crash lldb."""
        self.runCmd("help memory read", check=False)

    @no_debug_info_test
    def test_help_should_not_hang_emacsshell(self):
        """Command 'settings set term-width 0' should not hang the help command."""
        self.expect(
            "settings set term-width 0",
            COMMAND_FAILED_AS_EXPECTED,
            error=True,
            substrs=["error: 0 is out of range, valid values must be between"],
        )
        # self.runCmd("settings set term-width 0")
        self.expect("help", startstr="Debugger commands:")

    @no_debug_info_test
    def test_help_breakpoint_set(self):
        """Test that 'help breakpoint set' does not print out redundant lines of:
        'breakpoint set [-s <shlib-name>] ...'."""
        self.expect(
            "help breakpoint set",
            matching=False,
            substrs=["breakpoint set [-s <shlib-name>]"],
        )

    @no_debug_info_test
    def test_help_image_dump_symtab_should_not_crash(self):
        """Command 'help image dump symtab' should not crash lldb."""
        # 'image' is an alias for 'target modules'.
        self.expect("help image dump symtab", substrs=["dump symtab", "sort-order"])

    @no_debug_info_test
    def test_help_image_du_sym_is_ambiguous(self):
        """Command 'help image du sym' is ambiguous and spits out the list of candidates."""
        self.expect(
            "help image du sym",
            COMMAND_FAILED_AS_EXPECTED,
            error=True,
            substrs=["error: ambiguous command image du sym", "symfile", "symtab"],
        )

    @no_debug_info_test
    def test_help_image_du_line_should_work(self):
        """Command 'help image du line-table' is not ambiguous and should work."""
        # 'image' is an alias for 'target modules'.
        self.expect(
            "help image du line",
            substrs=["Dump the line table for one or more compilation units"],
        )

    @no_debug_info_test
    def test_help_image_list_shows_positional_args(self):
        """Command 'help image list' should describe positional args."""
        # 'image' is an alias for 'target modules'.
        self.expect("help image list", substrs=["<module> [...]"])

    @no_debug_info_test
    def test_help_target_variable_syntax(self):
        """Command 'help target variable' should display <variable-name> ..."""
        self.expect(
            "help target variable", substrs=["<variable-name> [<variable-name> [...]]"]
        )

    @no_debug_info_test
    def test_help_watchpoint_and_its_args(self):
        """Command 'help watchpoint', 'help watchpt-id', and 'help watchpt-id-list' should work."""
        self.expect("help watchpoint", substrs=["delete", "disable", "enable", "list"])
        self.expect("help watchpt-id", substrs=["<watchpt-id>"])
        self.expect("help watchpt-id-list", substrs=["<watchpt-id-list>"])

    @no_debug_info_test
    def test_help_watchpoint_set(self):
        """Test that 'help watchpoint set' prints out 'expression' and 'variable'
        as the possible subcommands."""
        self.expect(
            "help watchpoint set",
            substrs=["The following subcommands are supported:"],
            patterns=["expression +--", "variable +--"],
        )

    @no_debug_info_test
    def test_help_po_hides_options(self):
        """Test that 'help po' does not show all the options for expression"""
        self.expect(
            "help po",
            substrs=["--show-all-children", "--object-description"],
            matching=False,
        )

    @no_debug_info_test
    def test_help_run_hides_options(self):
        """Test that 'help run' does not show all the options for process launch"""
        self.expect("help run", substrs=["--arch", "--environment"], matching=False)

    @no_debug_info_test
    def test_help_next_shows_options(self):
        """Test that 'help next' shows all the options for thread step-over"""
        self.expect(
            "help next",
            substrs=["--step-out-avoids-no-debug", "--run-mode"],
            matching=True,
        )

    @no_debug_info_test
    def test_help_provides_alternatives(self):
        """Test that help on commands that don't exist provides information on additional help avenues"""
        self.expect(
            "help thisisnotadebuggercommand",
            substrs=[
                "'thisisnotadebuggercommand' is not a known command.",
                "Try 'help' to see a current list of commands.",
                "Try 'apropos thisisnotadebuggercommand' for a list of related commands.",
                "Try 'type lookup thisisnotadebuggercommand' for information on types, methods, functions, modules, etc.",
            ],
            error=True,
        )

        self.expect(
            "help process thisisnotadebuggercommand",
            substrs=[
                "'process thisisnotadebuggercommand' is not a known command.",
                "Try 'help' to see a current list of commands.",
                "Try 'apropos thisisnotadebuggercommand' for a list of related commands.",
                "Try 'type lookup thisisnotadebuggercommand' for information on types, methods, functions, modules, etc.",
            ],
        )

    @no_debug_info_test
    def test_custom_help_alias(self):
        """Test that aliases pick up custom help text."""

        def cleanup():
            self.runCmd("command unalias afriendlyalias", check=False)
            self.runCmd("command unalias averyfriendlyalias", check=False)

        self.addTearDownHook(cleanup)
        self.runCmd(
            'command alias --help "I am a friendly alias" -- afriendlyalias help'
        )
        self.expect(
            "help afriendlyalias", matching=True, substrs=["I am a friendly alias"]
        )
        self.runCmd(
            'command alias --long-help "I am a very friendly alias" -- averyfriendlyalias help'
        )
        self.expect(
            "help averyfriendlyalias",
            matching=True,
            substrs=["I am a very friendly alias"],
        )

    @no_debug_info_test
    def test_alias_prints_origin(self):
        """Test that 'help <unique_match_to_alias>' prints the alias origin."""

        def cleanup():
            self.runCmd("command unalias alongaliasname", check=False)

        self.addTearDownHook(cleanup)
        self.runCmd("command alias alongaliasname help")
        self.expect(
            "help alongaliasna",
            matching=True,
            substrs=["'alongaliasna' is an abbreviation for 'help'"],
        )

    @no_debug_info_test
    def test_hidden_help(self):
        self.expect("help -h", substrs=["_regexp-bt"])

    @no_debug_info_test
    def test_help_ambiguous(self):
        self.expect(
            "help g",
            substrs=[
                "Help requested with ambiguous command name, possible completions:",
                "gdb-remote",
                "gui",
            ],
        )

    @no_debug_info_test
    def test_help_unknown_flag(self):
        self.expect("help -z", error=True, substrs=["unknown or ambiguous option"])

    @no_debug_info_test
    def test_help_format_output(self):
        """Test that help output reaches TerminalWidth and wraps to the next
        line if needed."""
        self.runCmd("settings set term-width 105")
        self.expect(
            "help format",
            matching=True,
            patterns=[
                r"^<format> -- One of the format names \(or one-character names\) that can be used to show a variable's value:\n"
                r"\s+\"default\"\n"
            ],
        )

        # The length of the first line will not be exactly 105 because we split
        # at the last whitespace point before the limit.
        self.runCmd("settings set term-width 104")
        self.expect(
            "help format",
            matching=True,
            patterns=[
                r"^<format> -- One of the format names \(or one-character names\) that can be used to show a variable's\n"
                r"\s+value:\n"
            ],
        )

        self.runCmd("settings set term-width 90")
        self.expect(
            "help format",
            matching=True,
            patterns=[
                r"<format> -- One of the format names \(or one-character names\) that can be used to show a\n"
                r"\s+variable's value:\n"
            ],
        )

        # Check that line splitting works with newline characters too. The raw
        # input after the word "pointer" does not contain spaces among more than
        # a hundred subsequent characters, the words are separated by newlines.
        self.runCmd("settings set term-width 80")
        self.expect(
            "help format",
            matching=True,
            substrs=["'p' or \"pointer\""],
        )

    @no_debug_info_test
    def test_help_option_group_format_options_usage(self):
        """Test that help on commands that use OptionGroupFormat options provide relevant help specific to that command."""
        self.expect(
            "help memory read",
            matching=True,
            substrs=[
                "-f <format> ( --format <format> )",
                "Specify a format to be used for display.",
                "-s <byte-size> ( --size <byte-size> )",
                "The size in bytes to use when displaying with the selected format.",
            ],
        )

        self.expect(
            "help memory write",
            matching=True,
            substrs=[
                "-f <format> ( --format <format> )",
                "The format to use for each of the value to be written.",
                "-s <byte-size> ( --size <byte-size> )",
                "The size in bytes to write from input file or each value.",
            ],
        )

    @no_debug_info_test
    def test_help_option_description_terminal_width_no_ansi(self):
        """Test that help on commands formats option descriptions acccording
        to the terminal width."""
        # Should fit on one line.
        self.runCmd("settings set term-width 138")
        self.expect(
            "help breakpoint set",
            matching=True,
            patterns=[
                r"\s+Set the breakpoint only in this shared library.  Can repeat this option multiple times to specify multiple shared libraries.\n"
            ],
        )

        # Must be printed on two lines.
        self.runCmd("settings set term-width 100")
        self.expect(
            "help breakpoint set",
            matching=True,
            patterns=[
                r"\s+Set the breakpoint only in this shared library.  Can repeat this option multiple times\n"
                r"\s+to specify multiple shared libraries.\n"
            ],
        )

    @no_debug_info_test
    def test_help_option_description_terminal_width_with_ansi(self):
        """Test that help on commands formats option descriptions that include
        ANSI codes acccording to the terminal width."""
        # Note that because color is enabled, we will use ANSI cursor codes to
        # indent, rather than spaces.
        self.runCmd("settings set use-color on")

        # Should fit on one line.
        self.runCmd("settings set term-width 138")
        self.expect(
            "help breakpoint set",
            matching=True,
            substrs=[
                # The "S" of "Set" is underlined.
                "\x1b[12C\x1b[4mS\x1b[0met the breakpoint only in this shared library.  Can repeat this option multiple times to specify multiple shared libraries.\n"
            ],
        )

        self.runCmd("settings set term-width 100")
        self.expect(
            "help breakpoint set",
            matching=True,
            substrs=[
                "\x1b[12C\x1b[4mS\x1b[0met the breakpoint only in this shared library.  Can repeat this option multiple times\n"
                "\x1b[12Cto specify multiple shared libraries.\n"
            ],
        )

        # If we do not account for the difference between the visible character's
        # position and that character's real position into the string with the invisible
        # ANSI codes, we will crash in various ways. Writing tests for each width
        # would require duplicating the line splitting algorithm here. So instead,
        # we will try to provoke crashes if any exist, checking that the start
        # and end of the output is shown.
        for width in range(70, 150):
            self.runCmd(f"settings set term-width {width}")
            self.expect(
                "help breakpoint set", substrs=["\x1b[4mS\x1b[0met the", "libraries."]
            )

    @no_debug_info_test
    def test_help_shows_optional_short_options(self):
        """Test that optional short options are printed and that they are in
        alphabetical order with upper case options first."""
        self.expect(
            "help memory read", substrs=["memory read [-br]", "memory read [-AFLORTr]"]
        )
        self.expect(
            "help target modules lookup", substrs=["target modules lookup [-Airv]"]
        )

    @no_debug_info_test
    def test_help_shows_command_options_usage(self):
        """Test that we start the usage section with a specific line."""
        self.expect(
            "help memory read", substrs=["Command Options Usage:\n  memory read"]
        )

    @no_debug_info_test
    def test_help_detailed_information_spacing(self):
        """Test that we put a break between the usage and the options help lines,
        and between the options themselves."""
        self.expect(
            "help memory read",
            substrs=[
                "[<address-expression>]\n\n       -A ( --show-all-children )",
                # Starts with the end of the show-all-children line
                "to show.\n\n       -D",
            ],
        )

    @no_debug_info_test
    def test_help_detailed_information_ordering(self):
        """Test that we order options alphabetically, upper case first."""
        # You could test this with a simple regex like:
        # <upper case>.*<later upper case>.*<lower case>.*<later lower case>
        # Except that that could pass sometimes even with shuffled output.
        # This makes sure that doesn't happen.

        self.runCmd("help memory read")
        got = self.res.GetOutput()
        _, options_lines = got.split("Command Options Usage:")
        options_lines = options_lines.lstrip().splitlines()

        # Skip over "memory read [-xyz] lines.
        while "memory read" in options_lines[0]:
            options_lines.pop(0)
        # Plus the newline after that.
        options_lines.pop(0)

        short_options = []
        for line in options_lines:
            # Ignore line breaks and descriptions.
            # (not stripping the line here in case some line of the descriptions
            # happens to start with "-")
            if not line or not line.startswith("       -"):
                continue
            # This apears at the end after the options.
            if "This command takes options and free form arguments." in line:
                break
            line = line.strip()
            # The order of -- only options is not enforced so ignore their position.
            if not line.startswith("--"):
                # Save its short char name.
                short_options.append(line[1])

        self.assertEqual(
            sorted(short_options),
            short_options,
            "Short option help displayed in an incorrect order!",
        )

    @no_debug_info_test
    def test_help_show_tags(self):
        """Check that memory find and memory read have the --show-tags option
        but only memory read mentions binary output."""
        self.expect(
            "help memory read",
            patterns=[
                "--show-tags\n\\s+Include memory tags in output "
                "\\(does not apply to binary output\\)."
            ],
        )
        self.expect(
            "help memory find",
            patterns=["--show-tags\n\\s+Include memory tags in output."],
        )

    @no_debug_info_test
    def test_help_show_enum_values(self):
        """Check the help output for a argument type contains the enum values
        and their descriptions."""
        self.expect(
            "help <log-handler>",
            substrs=[
                "The log handle that will be used to write out log messages.",
                "default",
                "Use the default (stream) log handler",
                "stream",
                "Write log messages to the debugger output stream",
                "circular",
                "Write log messages to a fixed size circular buffer",
                "os",
                "Write log messages to the operating system log",
            ],
        )



# RUN: %PYTHON %s | FileCheck %s

import gc
import io
import itertools
from mlir.ir import *


def run(f):
    print("\nTEST:", f.__name__)
    f()
    gc.collect()
    assert Context._get_live_count() == 0
    return f


# CHECK-LABEL: TEST: testSymbolTableInsert
@run
def testSymbolTableInsert():
    with Context() as ctx:
        ctx.allow_unregistered_dialects = True
        m1 = Module.parse(
            """
      func.func private @foo()
      func.func private @bar()"""
        )
        m2 = Module.parse(
            """
      func.func private @qux()
      func.func private @foo()
      "foo.bar"() : () -> ()"""
        )

        symbol_table = SymbolTable(m1.operation)

        # CHECK: func private @foo
        # CHECK: func private @bar
        assert "foo" in symbol_table
        print(symbol_table["foo"])
        assert "bar" in symbol_table
        bar = symbol_table["bar"]
        print(symbol_table["bar"])

        assert "qux" not in symbol_table

        del symbol_table["bar"]
        try:
            symbol_table.erase(symbol_table["bar"])
        except KeyError:
            pass
        else:
            assert False, "expected KeyError"

        # CHECK: module
        # CHECK:   func private @foo()
        print(m1)
        assert "bar" not in symbol_table

        bar._set_invalid()
        try:
            print(bar)
        except RuntimeError as e:
            if "the operation has been invalidated" not in str(e):
                raise
        else:
            assert False, "expected RuntimeError due to invalidated operation"

        qux = m2.body.operations[0]
        m1.body.append(qux)
        symbol_table.insert(qux)
        assert "qux" in symbol_table

        # Check that insertion actually renames this symbol in the symbol table.
        foo2 = m2.body.operations[0]
        m1.body.append(foo2)
        updated_name = symbol_table.insert(foo2)
        assert foo2.name.value != "foo"
        assert foo2.name == updated_name
        assert isinstance(updated_name, StringAttr)

        # CHECK: module
        # CHECK:   func private @foo()
        # CHECK:   func private @qux()
        # CHECK:   func private @foo{{.*}}
        print(m1)

        try:
            symbol_table.insert(m2.body.operations[0])
        except ValueError as e:
            if "Expected operation to have a symbol name" not in str(e):
                raise
        else:
            assert False, "exepcted ValueError when adding a non-symbol"


# CHECK-LABEL: testSymbolTableRAUW
@run
def testSymbolTableRAUW():
    with Context() as ctx:
        m = Module.parse(
            """
      func.func private @foo() {
        call @bar() : () -> ()
        return
      }
      func.func private @bar()
      """
        )
        foo, bar = list(m.operation.regions[0].blocks[0].operations)[0:2]

        # Do renaming just within `foo`.
        SymbolTable.set_symbol_name(bar, "bam")
        SymbolTable.replace_all_symbol_uses("bar", "bam", foo)
        # CHECK: call @bam()
        # CHECK: func private @bam
        print(m)
        # CHECK: Foo symbol: StringAttr("foo")
        # CHECK: Bar symbol: StringAttr("bam")
        print(f"Foo symbol: {repr(SymbolTable.get_symbol_name(foo))}")
        print(f"Bar symbol: {repr(SymbolTable.get_symbol_name(bar))}")

        # Do renaming within the module.
        SymbolTable.set_symbol_name(bar, "baz")
        SymbolTable.replace_all_symbol_uses("bam", "baz", m.operation)
        # CHECK: call @baz()
        # CHECK: func private @baz
        print(m)
        # CHECK: Foo symbol: StringAttr("foo")
        # CHECK: Bar symbol: StringAttr("baz")
        print(f"Foo symbol: {repr(SymbolTable.get_symbol_name(foo))}")
        print(f"Bar symbol: {repr(SymbolTable.get_symbol_name(bar))}")


# CHECK-LABEL: testSymbolTableVisibility
@run
def testSymbolTableVisibility():
    with Context() as ctx:
        m = Module.parse(
            """
      func.func private @foo() {
        return
      }
      """
        )
        foo = m.operation.regions[0].blocks[0].operations[0]
        # CHECK: Existing visibility: StringAttr("private")
        print(f"Existing visibility: {repr(SymbolTable.get_visibility(foo))}")
        SymbolTable.set_visibility(foo, "public")
        # CHECK: func public @foo
        print(m)


# CHECK: testWalkSymbolTables
@run
def testWalkSymbolTables():
    with Context() as ctx:
        m = Module.parse(
            """
      module @outer {
        module @inner{
        }
      }
      """
        )

        def callback(symbol_table_op, uses_visible):
            print(f"SYMBOL TABLE: {uses_visible}: {symbol_table_op}")

        # CHECK: SYMBOL TABLE: True: module @inner
        # CHECK: SYMBOL TABLE: True: module @outer
        SymbolTable.walk_symbol_tables(m.operation, True, callback)

        # Make sure exceptions in the callback are handled.
        def error_callback(symbol_table_op, uses_visible):
            assert False, "Raised from python"

        try:
            SymbolTable.walk_symbol_tables(m.operation, True, error_callback)
        except RuntimeError as e:
            # CHECK: GOT EXCEPTION: Exception raised in callback:
            # CHECK: AssertionError: Raised from python
            print(f"GOT EXCEPTION: {e}")



"""
Test 'watchpoint command'.
"""


import os
import lldb
from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *
from lldbsuite.test import lldbutil


class WatchpointPythonCommandTestCase(TestBase):
    SHARED_BUILD_TESTCASE = False
    NO_DEBUG_INFO_TESTCASE = True

    def setUp(self):
        # Call super's setUp().
        TestBase.setUp(self)
        # Our simple source filename.
        self.source = "main.cpp"
        # Find the line number to break inside main().
        self.line = line_number(self.source, "// Set break point at this line.")
        self.second_line = line_number(self.source, "// Set another breakpoint here.")
        # And the watchpoint variable declaration line number.
        self.decl = line_number(self.source, "// Watchpoint variable declaration.")
        # Build dictionary to have unique executable names for each test
        # method.
        self.exe_name = self.testMethodName
        self.d = {"CXX_SOURCES": self.source, "EXE": self.exe_name}

    def test_watchpoint_command(self):
        """Test 'watchpoint command'."""
        self.build(dictionary=self.d)
        self.setTearDownCleanup(dictionary=self.d)

        exe = self.getBuildArtifact(self.exe_name)
        self.runCmd("file " + exe, CURRENT_EXECUTABLE_SET)

        # Add a breakpoint to set a watchpoint when stopped on the breakpoint.
        lldbutil.run_break_set_by_file_and_line(
            self, None, self.line, num_expected_locations=1
        )

        # Run the program.
        self.runCmd("run", RUN_SUCCEEDED)

        # We should be stopped again due to the breakpoint.
        # The stop reason of the thread should be breakpoint.
        self.expect(
            "thread list",
            STOPPED_DUE_TO_BREAKPOINT,
            substrs=["stopped", "stop reason = breakpoint"],
        )

        # Now let's set a write-type watchpoint for 'global'.
        self.expect(
            "watchpoint set variable -w write global",
            WATCHPOINT_CREATED,
            substrs=[
                "Watchpoint created",
                "size = 4",
                "type = w",
                "%s:%d" % (self.source, self.decl),
            ],
        )

        self.runCmd(
            "watchpoint command add -s python 1 -o 'frame.EvaluateExpression(\"cookie = 777\")'"
        )

        # List the watchpoint command we just added.
        self.expect(
            "watchpoint command list 1",
            substrs=["frame.EvaluateExpression", "cookie = 777"],
        )

        # Use the '-v' option to do verbose listing of the watchpoint.
        # The hit count should be 0 initially.
        self.expect("watchpoint list -v", substrs=["hit_count = 0"])

        self.runCmd("process continue")

        # We should be stopped again due to the watchpoint (write type).
        # The stop reason of the thread should be watchpoint.
        self.expect(
            "thread backtrace",
            STOPPED_DUE_TO_WATCHPOINT,
            substrs=["stop reason = watchpoint"],
        )

        # Check that the watchpoint snapshoting mechanism is working.
        self.expect("watchpoint list -v", substrs=["old value: 0", "new value: 1"])

        # The watchpoint command "forced" our global variable 'cookie' to
        # become 777.
        self.expect(
            "frame variable --show-globals cookie",
            substrs=["(int32_t)", "cookie = 777"],
        )

    def test_continue_in_watchpoint_command(self):
        """Test continue in a watchpoint command."""
        self.build(dictionary=self.d)
        self.setTearDownCleanup(dictionary=self.d)

        exe = self.getBuildArtifact(self.exe_name)
        self.runCmd("file " + exe, CURRENT_EXECUTABLE_SET)

        # Add a breakpoint to set a watchpoint when stopped on the breakpoint.
        lldbutil.run_break_set_by_file_and_line(
            self, None, self.line, num_expected_locations=1
        )

        # Run the program.
        self.runCmd("run", RUN_SUCCEEDED)

        # We should be stopped again due to the breakpoint.
        # The stop reason of the thread should be breakpoint.
        self.expect(
            "thread list",
            STOPPED_DUE_TO_BREAKPOINT,
            substrs=["stopped", "stop reason = breakpoint"],
        )

        # Now let's set a write-type watchpoint for 'global'.
        self.expect(
            "watchpoint set variable -w write global",
            WATCHPOINT_CREATED,
            substrs=[
                "Watchpoint created",
                "size = 4",
                "type = w",
                "%s:%d" % (self.source, self.decl),
            ],
        )

        cmd_script_file = os.path.join(self.getSourceDir(), "watchpoint_command.py")
        self.runCmd("command script import '%s'" % (cmd_script_file))

        self.runCmd("watchpoint command add -F watchpoint_command.watchpoint_command")

        # List the watchpoint command we just added.
        self.expect(
            "watchpoint command list 1",
            substrs=["watchpoint_command.watchpoint_command"],
        )

        self.runCmd("process continue")

        # We should be stopped again due to the watchpoint (write type).
        # The stop reason of the thread should be watchpoint.
        self.expect(
            "thread backtrace",
            STOPPED_DUE_TO_WATCHPOINT,
            substrs=["stop reason = watchpoint"],
        )

        # We should have hit the watchpoint once, set cookie to 888, since the
        # user callback returned True.
        self.expect(
            "frame variable --show-globals cookie",
            substrs=["(int32_t)", "cookie = 888"],
        )

        self.runCmd("process continue")

        # We should be stopped again due to the watchpoint (write type).
        # The stop reason of the thread should be watchpoint.
        self.expect(
            "thread backtrace",
            STOPPED_DUE_TO_WATCHPOINT,
            substrs=["stop reason = watchpoint"],
        )

        # We should have hit the watchpoint a second time, set cookie to 666,
        # even if the user callback didn't return anything and then continue.
        self.expect(
            "frame variable --show-globals cookie",
            substrs=["(int32_t)", "cookie = 666"],
        )

        # Add a breakpoint to set a watchpoint when stopped on the breakpoint.
        lldbutil.run_break_set_by_file_and_line(
            self, None, self.second_line, num_expected_locations=1
        )

        self.runCmd("process continue")

        self.expect(
            "thread backtrace",
            STOPPED_DUE_TO_BREAKPOINT,
            substrs=["stop reason = breakpoint"],
        )

        # We should have hit the watchpoint once, set cookie to 888, then continued to the
        # second hit and set it to 999
        self.expect(
            "frame variable --show-globals cookie",
            substrs=["(int32_t)", "cookie = 999"],
        )



import lldb
from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *
from lldbsuite.test import lldbutil


class TestCase(TestBase):
    @no_debug_info_test
    @skipIfWindows
    def test(self):
        self.build()
        lldbutil.run_to_source_breakpoint(
            self, "// break here", lldb.SBFileSpec("main.c")
        )

        self.expect_expr(
            "complex_float", result_type="_Complex float", result_value="-1.5 + -2.5i"
        )
        self.expect_expr(
            "complex_float + (2.0f + 3.5fi)",
            result_type="_Complex float",
            result_value="0.5 + 1i",
        )

        self.expect_expr(
            "complex_double", result_type="_Complex double", result_value="-1.5 + -2.5i"
        )
        self.expect_expr(
            "complex_double + (2.0 + 3.5i)",
            result_type="_Complex double",
            result_value="0.5 + 1i",
        )

    @no_debug_info_test
    # FIXME: LLDB fails to read the imaginary part of the number.
    @expectedFailureAll()
    @skipIfWindows
    def test_long_double(self):
        self.build()
        lldbutil.run_to_source_breakpoint(
            self, "// break here", lldb.SBFileSpec("main.c")
        )

        self.expect_expr(
            "complex_long_double",
            result_type="_Complex long double",
            result_value="-1.5 + 1i",
        )
        self.expect_expr(
            "complex_long_double + (2.0L + 3.5Li)",
            result_type="_Complex long double",
            result_value="0.5 + 1i",
        )



"""
Test SBType and SBTypeList API.
"""

import lldb
from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *
from lldbsuite.test import lldbutil
from lldbsuite.test import lldbplatformutil

class TypeAndTypeListTestCase(TestBase):
    SHARED_BUILD_TESTCASE = False
    def setUp(self):
        # Call super's setUp().
        TestBase.setUp(self)
        # We'll use the test method name as the exe_name.
        self.exe_name = self.testMethodName
        # Find the line number to break at.
        self.source = "main.cpp"
        self.line = line_number(self.source, "// Break at this line")

    def _find_nested_type_in_Pointer_template_arg(self, pointer_type):
        self.assertTrue(pointer_type)
        self.DebugSBType(pointer_type)
        pointer_info_type = pointer_type.template_args[1]
        self.assertTrue(pointer_info_type)
        self.DebugSBType(pointer_info_type)

        pointer_masks1_type = pointer_info_type.FindDirectNestedType("Masks1")
        self.assertTrue(pointer_masks1_type)
        self.DebugSBType(pointer_masks1_type)

        pointer_masks2_type = pointer_info_type.FindDirectNestedType("Masks2")
        self.assertTrue(pointer_masks2_type)
        self.DebugSBType(pointer_masks2_type)

    def _find_static_field_in_Task_pointer(self, task_pointer):
        self.assertTrue(task_pointer)
        self.DebugSBType(task_pointer)

        task_type = task_pointer.GetPointeeType()
        self.assertTrue(task_type)
        self.DebugSBType(task_type)

        static_constexpr_field = task_type.GetStaticFieldWithName(
            "static_constexpr_field"
        )
        self.assertTrue(static_constexpr_field)
        self.assertEqual(static_constexpr_field.GetName(), "static_constexpr_field")
        self.assertEqual(static_constexpr_field.GetType().GetName(), "const long")

        value = static_constexpr_field.GetConstantValue(self.target())
        self.DebugSBValue(value)
        self.assertEqual(value.GetValueAsSigned(), 47)

        static_constexpr_bool_field = task_type.GetStaticFieldWithName(
            "static_constexpr_bool_field"
        )
        self.assertTrue(static_constexpr_bool_field)
        self.assertEqual(
            static_constexpr_bool_field.GetName(), "static_constexpr_bool_field"
        )
        self.assertEqual(static_constexpr_bool_field.GetType().GetName(), "const bool")

        value = static_constexpr_bool_field.GetConstantValue(self.target())
        self.DebugSBValue(value)
        self.assertEqual(value.GetValueAsUnsigned(), 1)

        static_mutable_field = task_type.GetStaticFieldWithName("static_mutable_field")
        self.assertTrue(static_mutable_field)
        self.assertEqual(static_mutable_field.GetName(), "static_mutable_field")
        self.assertEqual(static_mutable_field.GetType().GetName(), "int")

        self.assertFalse(static_mutable_field.GetConstantValue(self.target()))

    @skipIf(compiler="clang", compiler_version=["<", "17.0"])
    def test(self):
        """Exercise SBType and SBTypeList API."""
        d = {"EXE": self.exe_name}
        self.build(dictionary=d)
        self.setTearDownCleanup(dictionary=d)
        exe = self.getBuildArtifact(self.exe_name)

        # Create a target by the debugger.
        target = self.dbg.CreateTarget(exe)
        self.assertTrue(target, VALID_TARGET)

        # Create the breakpoint inside function 'main'.
        breakpoint = target.BreakpointCreateByLocation(self.source, self.line)
        self.assertTrue(breakpoint, VALID_BREAKPOINT)

        # Now launch the process, and do not stop at entry point.
        process = target.LaunchSimple(None, None, self.get_process_working_directory())
        self.assertTrue(process, PROCESS_IS_VALID)

        # Get Frame #0.
        self.assertState(process.GetState(), lldb.eStateStopped)
        thread = lldbutil.get_stopped_thread(process, lldb.eStopReasonBreakpoint)
        self.assertTrue(
            thread.IsValid(),
            "There should be a thread stopped due to breakpoint condition",
        )
        frame0 = thread.GetFrameAtIndex(0)

        # Get the type 'Task'.
        type_list = target.FindTypes("Task")
        if self.TraceOn():
            print(
                "Size of type_list from target.FindTypes('Task') query: %d"
                % type_list.GetSize()
            )
        # a second Task make be scared up by the Objective-C runtime
        self.assertGreaterEqual(len(type_list), 1)
        self.assertEqual(
            type_list[0].GetName(),
            type_list.GetTypeAtIndex(0).GetName(),
            "subscript [0] matches GetTypeAtIndex(0)",
        )
        self.assertEqual(
            type_list[-1].GetName(),
            type_list.GetTypeAtIndex(type_list.GetSize() - 1).GetName(),
            "subscript [-1] matches last item",
        )
        for type in type_list:
            self.assertTrue(type)
            self.DebugSBType(type)
            self.assertFalse(type.IsAnonymousType(), "Task is not anonymous")
            self.assertTrue(type.IsAggregateType(), "Task is aggregate")
            for field in type.fields:
                if field.name == "type":
                    for enum_member in field.type.enum_members:
                        self.assertTrue(enum_member)
                        self.DebugSBType(enum_member.type)
                elif field.name == "my_type_is_nameless":
                    self.assertFalse(
                        field.type.IsAnonymousType(),
                        "my_type_is_nameless is not an anonymous type",
                    )
                    self.assertTrue(field.type.IsAggregateType())
                elif field.name == "my_type_is_named":
                    self.assertFalse(
                        field.type.IsAnonymousType(),
                        "my_type_is_named has a named type",
                    )
                    self.assertTrue(field.type.IsAggregateType())
                elif field.name is None:
                    self.assertTrue(
                        field.type.IsAnonymousType(), "Nameless type is not anonymous"
                    )

        # Pass an empty string.  LLDB should not crash. :-)
        fuzz_types = target.FindTypes(None)
        fuzz_type = target.FindFirstType(None)

        # Now use the SBTarget.FindFirstType() API to find 'Task'.
        task_type = target.FindFirstType("Task")
        self.assertTrue(task_type)
        self.DebugSBType(task_type)

        # Get the reference type of 'Task', just for fun.
        task_ref_type = task_type.GetReferenceType()
        self.assertTrue(task_ref_type)
        self.DebugSBType(task_ref_type)

        # Get the pointer type of 'Task', which is the same as task_head's
        # type.
        task_pointer_type = task_type.GetPointerType()
        self.assertTrue(task_pointer_type)
        self.DebugSBType(task_pointer_type)

        # Get variable 'task_head'.
        task_head = frame0.FindVariable("task_head")
        self.assertTrue(task_head, VALID_VARIABLE)
        self.DebugSBValue(task_head)
        task_head_type = task_head.GetType()
        self.DebugSBType(task_head_type)
        self.assertTrue(task_head_type.IsPointerType())
        self.assertFalse(task_head_type.IsAggregateType())

        self.assertEqual(task_head_type, task_pointer_type)

        # Get the pointee type of 'task_head'.
        task_head_pointee_type = task_head_type.GetPointeeType()
        self.DebugSBType(task_head_pointee_type)

        self.assertEqual(task_type, task_head_pointee_type)

        # Check whether we can find a directly nested type by name
        name_type = task_type.FindDirectNestedType("name")
        self.assertTrue(name_type)
        self.DebugSBType(name_type)

        enum_type = task_type.FindDirectNestedType("E")
        self.assertTrue(enum_type)
        self.DebugSBType(enum_type)

        union_type = task_type.FindDirectNestedType("U")
        self.assertTrue(union_type)
        self.DebugSBType(union_type)

        # Check that we don't find indirectly nested types
        self.assertEqual(enum_type.size, 1)

        invalid_type = task_type.FindDirectNestedType("E2")
        self.assertFalse(invalid_type)

        # Check that FindDirectNestedType handles types without DeclContext
        # and other errorneous inputs
        task_ptr_type = task_type.GetPointerType()
        invalid_type = task_ptr_type.FindDirectNestedType("name")
        self.assertFalse(invalid_type)

        invalid_type = task_type.FindDirectNestedType("")
        self.assertFalse(invalid_type)

        invalid_type = task_type.FindDirectNestedType(None)
        self.assertFalse(invalid_type)

        # Check that FindDirectNestedType works with types from module and
        # expression ASTs.
        self._find_nested_type_in_Pointer_template_arg(
            frame0.FindVariable("pointer").GetType()
        )
        self._find_nested_type_in_Pointer_template_arg(
            frame0.EvaluateExpression("pointer").GetType()
        )

        self._find_static_field_in_Task_pointer(
            frame0.FindVariable("task_head").GetType()
        )
        self._find_static_field_in_Task_pointer(
            frame0.EvaluateExpression("task_head").GetType()
        )

        # We'll now get the child member 'id' from 'task_head'.
        id = task_head.GetChildMemberWithName("id")
        self.DebugSBValue(id)
        id_type = id.GetType()
        self.DebugSBType(id_type)
        self.assertFalse(id_type.IsAggregateType())

        # SBType.GetBasicType() takes an enum 'BasicType'
        # (lldb-enumerations.h).
        int_type = id_type.GetBasicType(lldb.eBasicTypeInt)
        self.assertEqual(id_type, int_type)

        # Find 'myint_arr' and check the array element type.
        myint_arr = frame0.FindVariable("myint_arr")
        self.assertTrue(myint_arr, VALID_VARIABLE)
        self.DebugSBValue(myint_arr)
        myint_arr_type = myint_arr.GetType()
        self.DebugSBType(myint_arr_type)
        self.assertTrue(myint_arr_type.IsArrayType())
        self.assertTrue(myint_arr_type.IsAggregateType())
        myint_arr_element_type = myint_arr_type.GetArrayElementType()
        self.DebugSBType(myint_arr_element_type)
        myint_type = target.FindFirstType("myint")
        self.DebugSBType(myint_type)
        self.assertEqual(myint_arr_element_type, myint_type)

        # Test enum methods. Requires DW_AT_enum_class which was added in Dwarf 4.
        if int(lldbplatformutil.getDwarfVersion()) >= 4:
            enum_type = target.FindFirstType("EnumType")
            self.assertTrue(enum_type)
            self.DebugSBType(enum_type)
            self.assertFalse(enum_type.IsScopedEnumerationType())
            self.assertFalse(enum_type.IsAggregateType())

            scoped_enum_type = target.FindFirstType("ScopedEnumType")
            self.assertTrue(scoped_enum_type)
            self.DebugSBType(scoped_enum_type)
            self.assertTrue(scoped_enum_type.IsScopedEnumerationType())
            self.assertFalse(scoped_enum_type.IsAggregateType())
            int_scoped_enum_type = scoped_enum_type.GetEnumerationIntegerType()
            self.assertTrue(int_scoped_enum_type)
            self.DebugSBType(int_scoped_enum_type)
            self.assertEqual(int_scoped_enum_type.GetName(), "int")

            enum_uchar = target.FindFirstType("EnumUChar")
            self.assertTrue(enum_uchar)
            self.DebugSBType(enum_uchar)
            int_enum_uchar = enum_uchar.GetEnumerationIntegerType()
            self.assertTrue(int_enum_uchar)
            self.DebugSBType(int_enum_uchar)
            self.assertEqual(int_enum_uchar.GetName(), "unsigned char")

    def test_nested_typedef(self):
        """Exercise FindDirectNestedType for typedefs."""
        self.build()
        target = self.dbg.CreateTarget(self.getBuildArtifact())
        self.assertTrue(target)

        with_nested_typedef = target.FindFirstType("WithNestedTypedef")
        self.assertTrue(with_nested_typedef)

        # This is necessary to work around #91186
        self.assertTrue(target.FindFirstGlobalVariable("typedefed_value").GetType())

        the_typedef = with_nested_typedef.FindDirectNestedType("TheTypedef")
        self.assertTrue(the_typedef)
        self.assertEqual(the_typedef.GetTypedefedType().GetName(), "int")

    def test_GetByteAlign(self):
        """Exercise SBType::GetByteAlign"""
        self.build()
        spec = lldb.SBModuleSpec()
        spec.SetFileSpec(lldb.SBFileSpec(self.getBuildArtifact()))
        module = lldb.SBModule(spec)
        self.assertTrue(module)

        # Invalid types should not crash.
        self.assertEqual(lldb.SBType().GetByteAlign(), 0)

        # Try a type with natural alignment.
        void_ptr = module.GetBasicType(lldb.eBasicTypeVoid).GetPointerType()
        self.assertTrue(void_ptr)
        # Not exactly guaranteed by the spec, but should be true everywhere we
        # care about.
        self.assertEqual(void_ptr.GetByteSize(), void_ptr.GetByteAlign())

        # And an over-aligned type.
        self.assertEqual(module.FindFirstType("OverAlignedStruct").GetByteAlign(), 128)



"""
Test that lldb persistent types works correctly.
"""


import lldb
from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *
from lldbsuite.test import lldbutil


class PersistenttypesTestCase(TestBase):
    def test_persistent_types(self):
        """Test that lldb persistent types works correctly."""
        self.build()

        self.runCmd("file " + self.getBuildArtifact("a.out"), CURRENT_EXECUTABLE_SET)

        self.runCmd("breakpoint set --name main")

        self.runCmd("run", RUN_SUCCEEDED)

        self.runCmd("expression struct $foo { int a; int b; };")

        self.expect(
            "expression struct $foo $my_foo; $my_foo.a = 2; $my_foo.b = 3;",
            startstr="(int) $0 = 3",
        )

        self.expect("expression $my_foo", substrs=["a = 2", "b = 3"])

        self.runCmd("expression typedef int $bar")

        self.expect("expression $bar i = 5; i", startstr="($bar) $1 = 5")

        self.runCmd("expression struct $foobar { char a; char b; char c; char d; };")
        self.runCmd("next")

        self.expect(
            "memory read foo -t $foobar",
            substrs=["($foobar) 0x", " = ", "a = 'H'", "b = 'e'", "c = 'l'", "d = 'l'"],
        )  # persistent types are OK to use for memory read

        self.expect(
            "memory read foo -t $foobar -x c",
            substrs=["($foobar) 0x", " = ", "a = 'H'", "b = 'e'", "c = 'l'", "d = 'l'"],
        )  # persistent types are OK to use for memory read

        self.expect(
            "memory read foo -t foobar",
            substrs=["($foobar) 0x", " = ", "a = 'H'", "b = 'e'", "c = 'l'", "d = 'l'"],
            matching=False,
            error=True,
        )  # the type name is $foobar, make sure we settle for nothing less

        self.expect(
            "expression struct { int a; int b; } x = { 2, 3 }; x",
            substrs=["a = 2", "b = 3"],
        )

        self.expect(
            "expression struct { int x; int y; int z; } object; object.y = 1; object.z = 3; object.x = 2; object",
            substrs=["x = 2", "y = 1", "z = 3"],
        )

        self.expect(
            "expression struct A { int x; int y; }; struct { struct A a; int z; } object; object.a.y = 1; object.z = 3; object.a.x = 2; object",
            substrs=["x = 2", "y = 1", "z = 3"],
        )



"""
Test watchpoint list, enable, disable, and delete commands.
"""


import lldb
from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *
from lldbsuite.test import lldbutil


class WatchpointCommandsTestCase(TestBase):
    SHARED_BUILD_TESTCASE = False
    NO_DEBUG_INFO_TESTCASE = True

    def setUp(self):
        # Call super's setUp().
        TestBase.setUp(self)
        # Our simple source filename.
        self.source = "main.c"
        # Find the line number to break inside main().
        self.line = line_number(self.source, "// Set break point at this line.")
        self.line2 = line_number(
            self.source, "// Set 2nd break point for disable_then_enable test case."
        )
        # And the watchpoint variable declaration line number.
        self.decl = line_number(self.source, "// Watchpoint variable declaration.")
        # Build dictionary to have unique executable names for each test
        # method.
        self.exe_name = self.testMethodName
        self.d = {"C_SOURCES": self.source, "EXE": self.exe_name}

    # Read-write watchpoints not supported on SystemZ
    @expectedFailureAll(archs=["s390x"])
    def test_rw_watchpoint(self):
        """Test read_write watchpoint and expect to stop two times."""
        self.build(dictionary=self.d)
        self.setTearDownCleanup(dictionary=self.d)

        exe = self.getBuildArtifact(self.exe_name)
        self.runCmd("file " + exe, CURRENT_EXECUTABLE_SET)

        # Add a breakpoint to set a watchpoint when stopped on the breakpoint.
        lldbutil.run_break_set_by_file_and_line(
            self, None, self.line, num_expected_locations=1
        )

        # Run the program.
        self.runCmd("run", RUN_SUCCEEDED)

        # We should be stopped again due to the breakpoint.
        # The stop reason of the thread should be breakpoint.
        self.expect(
            "thread list",
            STOPPED_DUE_TO_BREAKPOINT,
            substrs=["stopped", "stop reason = breakpoint"],
        )

        # Now let's set a read_write-type watchpoint for 'global'.
        # There should be two watchpoint hits (see main.c).
        self.expect(
            "watchpoint set variable -w read_write global",
            WATCHPOINT_CREATED,
            substrs=[
                "Watchpoint created",
                "size = 4",
                "type = rw",
                "%s:%d" % (self.source, self.decl),
            ],
        )

        # Use the '-v' option to do verbose listing of the watchpoint.
        # The hit count should be 0 initially.
        self.expect(
            "watchpoint list -v",
            substrs=["Number of supported hardware watchpoints:", "hit_count = 0"],
        )

        self.runCmd("process continue")

        # We should be stopped again due to the watchpoint (read_write type).
        # The stop reason of the thread should be watchpoint.
        self.expect(
            "thread backtrace",
            STOPPED_DUE_TO_WATCHPOINT,
            substrs=["stop reason = watchpoint"],
        )

        self.runCmd("process continue")

        # We should be stopped again due to the watchpoint (read_write type).
        # The stop reason of the thread should be watchpoint.
        self.expect(
            "thread backtrace",
            STOPPED_DUE_TO_WATCHPOINT,
            substrs=["stop reason = watchpoint"],
        )

        self.runCmd("process continue")

        # There should be no more watchpoint hit and the process status should
        # be 'exited'.
        self.expect("process status", substrs=["exited"])

        # Use the '-v' option to do verbose listing of the watchpoint.
        # The hit count should now be 2.
        self.expect("watchpoint list -v", substrs=["hit_count = 2"])

    # Read-write watchpoints not supported on SystemZ
    @expectedFailureAll(archs=["s390x"])
    def test_rw_watchpoint_delete(self):
        """Test delete watchpoint and expect not to stop for watchpoint."""
        self.build()
        lldbutil.run_to_line_breakpoint(self, lldb.SBFileSpec(self.source), self.line)

        # Now let's set a read_write-type watchpoint for 'global'.
        # There should be two watchpoint hits (see main.c).
        self.expect(
            "watchpoint set variable -w read_write global",
            WATCHPOINT_CREATED,
            substrs=[
                "Watchpoint created",
                "size = 4",
                "type = rw",
                "%s:%d" % (self.source, self.decl),
            ],
        )

        # Delete the watchpoint immediately, but set auto-confirm to true
        # first.
        self.runCmd("settings set auto-confirm true")
        self.expect("watchpoint delete", substrs=["All watchpoints removed."])
        # Restore the original setting of auto-confirm.
        self.runCmd("settings clear auto-confirm")

        target = self.dbg.GetSelectedTarget()
        self.assertTrue(target and not target.GetNumWatchpoints())

        # Now let's set a read_write-type watchpoint for 'global'.
        # There should be two watchpoint hits (see main.c).
        self.expect(
            "watchpoint set variable -w read_write global",
            WATCHPOINT_CREATED,
            substrs=[
                "Watchpoint created",
                "size = 4",
                "type = rw",
                "%s:%d" % (self.source, self.decl),
            ],
        )

        # Delete the watchpoint immediately using the force option.
        self.expect("watchpoint delete --force", substrs=["All watchpoints removed."])

        self.assertTrue(target and not target.GetNumWatchpoints())

        self.runCmd("process continue")

        # There should be no more watchpoint hit and the process status should
        # be 'exited'.
        self.expect("process status", substrs=["exited"])

    # Read-write watchpoints not supported on SystemZ
    @expectedFailureAll(archs=["s390x"])
    def test_rw_watchpoint_set_ignore_count(self):
        """Test watchpoint ignore count and expect to not to stop at all."""
        self.build(dictionary=self.d)
        self.setTearDownCleanup(dictionary=self.d)

        exe = self.getBuildArtifact(self.exe_name)
        self.runCmd("file " + exe, CURRENT_EXECUTABLE_SET)

        # Add a breakpoint to set a watchpoint when stopped on the breakpoint.
        lldbutil.run_break_set_by_file_and_line(
            self, None, self.line, num_expected_locations=1
        )

        # Run the program.
        self.runCmd("run", RUN_SUCCEEDED)

        # We should be stopped again due to the breakpoint.
        # The stop reason of the thread should be breakpoint.
        self.expect(
            "thread list",
            STOPPED_DUE_TO_BREAKPOINT,
            substrs=["stopped", "stop reason = breakpoint"],
        )

        # Now let's set a read_write-type watchpoint for 'global'.
        # There should be two watchpoint hits (see main.c).
        self.expect(
            "watchpoint set variable -w read_write global",
            WATCHPOINT_CREATED,
            substrs=[
                "Watchpoint created",
                "size = 4",
                "type = rw",
                "%s:%d" % (self.source, self.decl),
            ],
        )

        # Set the ignore count of the watchpoint immediately.
        self.expect("watchpoint ignore -i 2", substrs=["All watchpoints ignored."])

        # Use the '-v' option to do verbose listing of the watchpoint.
        # Expect to find an ignore_count of 2.
        self.expect("watchpoint list -v", substrs=["hit_count = 0", "ignore_count = 2"])

        self.runCmd("process continue")

        # There should be no more watchpoint hit and the process status should
        # be 'exited'.
        self.expect("process status", substrs=["exited"])

        # Use the '-v' option to do verbose listing of the watchpoint.
        # Expect to find a hit_count of 2 as well.
        self.expect("watchpoint list -v", substrs=["hit_count = 2", "ignore_count = 2"])

    # Read-write watchpoints not supported on SystemZ
    @expectedFailureAll(archs=["s390x"])
    def test_rw_disable_after_first_stop(self):
        """Test read_write watchpoint but disable it after the first stop."""
        self.build(dictionary=self.d)
        self.setTearDownCleanup(dictionary=self.d)

        exe = self.getBuildArtifact(self.exe_name)
        self.runCmd("file " + exe, CURRENT_EXECUTABLE_SET)

        # Add a breakpoint to set a watchpoint when stopped on the breakpoint.
        lldbutil.run_break_set_by_file_and_line(
            self, None, self.line, num_expected_locations=1
        )

        # Run the program.
        self.runCmd("run", RUN_SUCCEEDED)

        # We should be stopped again due to the breakpoint.
        # The stop reason of the thread should be breakpoint.
        self.expect(
            "thread list",
            STOPPED_DUE_TO_BREAKPOINT,
            substrs=["stopped", "stop reason = breakpoint"],
        )

        # Now let's set a read_write-type watchpoint for 'global'.
        # There should be two watchpoint hits (see main.c).
        self.expect(
            "watchpoint set variable -w read_write global",
            WATCHPOINT_CREATED,
            substrs=[
                "Watchpoint created",
                "size = 4",
                "type = rw",
                "%s:%d" % (self.source, self.decl),
            ],
        )

        # Use the '-v' option to do verbose listing of the watchpoint.
        # The hit count should be 0 initially.
        self.expect("watchpoint list -v", substrs=["state = enabled", "hit_count = 0"])

        self.runCmd("process continue")

        # We should be stopped again due to the watchpoint (read_write type).
        # The stop reason of the thread should be watchpoint.
        self.expect(
            "thread backtrace",
            STOPPED_DUE_TO_WATCHPOINT,
            substrs=["stop reason = watchpoint"],
        )

        # Before continuing, we'll disable the watchpoint, which means we won't
        # stop again after this.
        self.runCmd("watchpoint disable")

        self.expect("watchpoint list -v", substrs=["state = disabled", "hit_count = 1"])

        self.runCmd("process continue")

        # There should be no more watchpoint hit and the process status should
        # be 'exited'.
        self.expect("process status", substrs=["exited"])

        # Use the '-v' option to do verbose listing of the watchpoint.
        # The hit count should be 1.
        self.expect("watchpoint list -v", substrs=["hit_count = 1"])

    # Read-write watchpoints not supported on SystemZ
    @expectedFailureAll(archs=["s390x"])
    def test_rw_disable_then_enable(self):
        """Test read_write watchpoint, disable initially, then enable it."""
        self.build(dictionary=self.d)
        self.setTearDownCleanup(dictionary=self.d)

        exe = self.getBuildArtifact(self.exe_name)
        self.runCmd("file " + exe, CURRENT_EXECUTABLE_SET)

        # Add a breakpoint to set a watchpoint when stopped on the breakpoint.
        lldbutil.run_break_set_by_file_and_line(
            self, None, self.line, num_expected_locations=1
        )
        lldbutil.run_break_set_by_file_and_line(
            self, None, self.line2, num_expected_locations=1
        )

        # Run the program.
        self.runCmd("run", RUN_SUCCEEDED)

        # We should be stopped again due to the breakpoint.
        # The stop reason of the thread should be breakpoint.
        self.expect(
            "thread list",
            STOPPED_DUE_TO_BREAKPOINT,
            substrs=["stopped", "stop reason = breakpoint"],
        )

        # Now let's set a read_write-type watchpoint for 'global'.
        # There should be two watchpoint hits (see main.c).
        self.expect(
            "watchpoint set variable -w read_write global",
            WATCHPOINT_CREATED,
            substrs=[
                "Watchpoint created",
                "size = 4",
                "type = rw",
                "%s:%d" % (self.source, self.decl),
            ],
        )

        # Immediately, we disable the watchpoint.  We won't be stopping due to a
        # watchpoint after this.
        self.runCmd("watchpoint disable")

        # Use the '-v' option to do verbose listing of the watchpoint.
        # The hit count should be 0 initially.
        self.expect("watchpoint list -v", substrs=["state = disabled", "hit_count = 0"])

        self.runCmd("process continue")

        # We should be stopped again due to the breakpoint.
        self.expect(
            "thread backtrace",
            STOPPED_DUE_TO_BREAKPOINT,
            substrs=["stop reason = breakpoint"],
        )

        # Before continuing, we'll enable the watchpoint, which means we will
        # stop again after this.
        self.runCmd("watchpoint enable")

        self.expect("watchpoint list -v", substrs=["state = enabled", "hit_count = 0"])

        self.runCmd("process continue")

        # We should be stopped again due to the watchpoint (read_write type).
        # The stop reason of the thread should be watchpoint.
        self.expect(
            "thread backtrace",
            STOPPED_DUE_TO_WATCHPOINT,
            substrs=["stop reason = watchpoint"],
        )

        self.runCmd("process continue")

        # There should be no more watchpoint hit and the process status should
        # be 'exited'.
        self.expect("process status", substrs=["exited"])

        # Use the '-v' option to do verbose listing of the watchpoint.
        # The hit count should be 1.
        self.expect("watchpoint list -v", substrs=["hit_count = 1"])



"""
Test defining commands using the lldb command definitions
"""
import inspect
import sys
import lldb
from lldb.plugins.parsed_cmd import ParsedCommand


class ReportingCmd(ParsedCommand):
    def __init__(self, debugger, unused):
        super().__init__(debugger, unused)

    def __call__(self, debugger, args_array, exe_ctx, result):
        opt_def = self.get_options_definition()
        if len(opt_def):
            result.AppendMessage("Options:\n")
            for long_option, elem in opt_def.items():
                if "value_type" in elem:
                    print(f"Looking at {long_option} - {elem}")
                    dest = elem["dest"]
                    result.AppendMessage(
                        f"{long_option} (set: {elem['_value_set']}): {object.__getattribute__(self.get_parser(), dest)}\n"
                    )
                else:
                    result.AppendMessage(
                        f"{long_option} (set: {elem['_value_set']}): flag value\n"
                    )
        else:
            result.AppendMessage("No options\n")

        num_args = args_array.GetSize()
        if num_args > 0:
            result.AppendMessage(f"{num_args} arguments:")
        for idx in range(0, num_args):
            result.AppendMessage(
                f"{idx}: {args_array.GetItemAtIndex(idx).GetStringValue(10000)}\n"
            )

# Use these to make sure that get_repeat_command sends the right
# command.
no_args_repeat = None
one_arg_repeat = None
two_arg_repeat = None

class NoArgsCommand(ReportingCmd):
    program = "no-args"

    def __init__(self, debugger, unused):
        super().__init__(debugger, unused)

    @classmethod
    def register_lldb_command(cls, debugger, module_name):
        ParsedCommand.do_register_cmd(cls, debugger, module_name)

    def setup_command_definition(self):
        ov_parser = self.get_parser()
        ov_parser.add_option(
            "b",
            "bool-arg",
            "a boolean arg, defaults to True",
            value_type=lldb.eArgTypeBoolean,
            groups=[1, 2],
            dest="bool_arg",
            default=True,
        )

        ov_parser.add_option(
            "s",
            "shlib-name",
            "A shared library name.",
            value_type=lldb.eArgTypeShlibName,
            groups=[1, [3, 4]],
            dest="shlib_name",
            default=None,
        )

        ov_parser.add_option(
            "d",
            "disk-file-name",
            "An on disk filename",
            value_type=lldb.eArgTypeFilename,
            dest="disk_file_name",
            default=None,
        )

        ov_parser.add_option("f", "flag-value", "This is a flag value")

        ov_parser.add_option(
            "l",
            "line-num",
            "A line number",
            value_type=lldb.eArgTypeLineNum,
            groups=3,
            dest="line_num",
            default=0,
        )

        ov_parser.add_option(
            "e",
            "enum-option",
            "An enum, doesn't actually do anything",
            enum_values=[
                ["foo", "does foo things"],
                ["bar", "does bar things"],
                ["baz", "does baz things"],
            ],
            groups=4,
            dest="enum_option",
            default="foo",
        )

    def get_repeat_command(self, command):
        # No auto-repeat
        global no_args_repeat
        no_args_repeat = command
        return ""

    def get_short_help(self):
        return "Example command for use in debugging"

    def get_long_help(self):
        return self.help_string


class OneArgCommandNoOptions(ReportingCmd):
    program = "one-arg-no-opt"

    def __init__(self, debugger, unused):
        super().__init__(debugger, unused)

    @classmethod
    def register_lldb_command(cls, debugger, module_name):
        ParsedCommand.do_register_cmd(cls, debugger, module_name)

    def setup_command_definition(self):
        ov_parser = self.get_parser()
        ov_parser.add_argument_set(
            [ov_parser.make_argument_element(lldb.eArgTypeSourceFile, "plain")]
        )

    def get_repeat_command(self, command):
        # Repeat the current command
        global one_arg_repeat
        one_arg_repeat = command
        return None

    def get_short_help(self):
        return "Example command for use in debugging"

    def get_long_help(self):
        return self.help_string


class TwoArgGroupsCommand(ReportingCmd):
    program = "two-args"

    def __init__(self, debugger, unused):
        super().__init__(debugger, unused)

    @classmethod
    def register_lldb_command(cls, debugger, module_name):
        ParsedCommand.do_register_cmd(cls, debugger, module_name)

    def setup_command_definition(self):
        ov_parser = self.get_parser()
        ov_parser.add_option(
            "l",
            "language",
            "language defaults to None",
            value_type=lldb.eArgTypeLanguage,
            groups=[1, 2],
            dest="language",
            default=None,
        )

        ov_parser.add_option(
            "c",
            "log-channel",
            "log channel - defaults to lldb",
            value_type=lldb.eArgTypeLogChannel,
            groups=[1, 3],
            dest="log_channel",
            default="lldb",
        )

        ov_parser.add_option(
            "p",
            "process-name",
            "A process name, defaults to None",
            value_type=lldb.eArgTypeProcessName,
            dest="proc_name",
            default=None,
        )

        ov_parser.add_argument_set(
            [
                ov_parser.make_argument_element(
                    lldb.eArgTypeClassName, "plain", [1, 2]
                ),
                ov_parser.make_argument_element(
                    lldb.eArgTypeOffset, "optional", [1, 2]
                ),
            ]
        )

        ov_parser.add_argument_set(
            [
                ov_parser.make_argument_element(
                    lldb.eArgTypePythonClass, "plain", [3, 4]
                ),
                ov_parser.make_argument_element(lldb.eArgTypePid, "optional", [3, 4]),
            ]
        )

    def get_repeat_command(self, command):
        global two_arg_repeat
        two_arg_repeat = command
        return command + " THIRD_ARG"

    def handle_option_argument_completion(self, long_option, cursor_pos):
        ov_parser = self.get_parser()
        value = ov_parser.dest_for_option(long_option)[0 : cursor_pos + 1]
        proc_value = ov_parser.proc_name
        if proc_value != None:
            new_str = value + proc_value
            ret_arr = {"completion": new_str, "mode": "partial"}
            return ret_arr

        ret_arr = {"values": [value + "nice", value + "not_nice", value + "mediocre"]}
        return ret_arr

    def handle_argument_completion(self, args, arg_pos, cursor_pos):
        ov_parser = self.get_parser()
        orig_arg = args[arg_pos][0:cursor_pos]
        if orig_arg == "correct_":
            ret_arr = {"completion": "correct_answer"}
            return ret_arr

        if ov_parser.was_set("process-name"):
            # No completions if proc_name was set.
            return True

        ret_arr = {
            "values": [orig_arg + "cool", orig_arg + "yuck"],
            "descriptions": ["good idea", "bad idea"],
        }
        return ret_arr

    def get_short_help(self):
        return "This is my short help string"

    def get_long_help(self):
        return self.help_string


def __lldb_init_module(debugger, dict):
    # Register all classes that have a register_lldb_command method
    for _name, cls in inspect.getmembers(sys.modules[__name__]):
        if inspect.isclass(cls) and callable(
            getattr(cls, "register_lldb_command", None)
        ):
            cls.register_lldb_command(debugger, __name__)



import lldb
from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *
from lldbsuite.test import lldbutil


class TestSBValueSynthetic(TestBase):
    NO_DEBUG_INFO_TESTCASE = True

    def test_str(self):
        self.build()
        lldbutil.run_to_source_breakpoint(
            self, "break here", lldb.SBFileSpec("main.cpp")
        )
        self.runCmd("command script import formatter.py")
        self.runCmd(
            "type synthetic add --python-class formatter.FooSyntheticProvider Foo"
        )

        formatted = self.frame().FindVariable("foo")
        has_formatted = self.frame().FindVariable("has_foo")
        self.expect(str(formatted), exe=False, substrs=["synth_child"])
        self.expect(str(has_formatted), exe=False, substrs=["synth_child"])

    def test_synth_arr(self):
        self.build()
        lldbutil.run_to_source_breakpoint(
            self, "break here", lldb.SBFileSpec("main.cpp")
        )
        point_arr = self.frame().FindVariable("point_arr")
        point_ptr = self.frame().FindVariable("point_ptr")
        for v in [point_arr, point_ptr]:
            for i in range(3):
                child = v.GetChildAtIndex(i, lldb.eDynamicDontRunTarget, True)
                check = ValueCheck(
                    name=f"[{i}]",
                    type="Point",
                    children=[
                        ValueCheck(name="x", value=str(2 * i + 1)),
                        ValueCheck(name="y", value=str(2 * i + 2)),
                    ],
                )
                check.check_value(self, child, f"{child}, child {i} of {v.GetName()}")

        int_arr = self.frame().FindVariable("int_arr")
        int_ptr = self.frame().FindVariable("int_ptr")
        for v in [int_arr, int_ptr]:
            for i in range(3):
                child = v.GetChildAtIndex(i, lldb.eDynamicDontRunTarget, True)
                check = ValueCheck(name=f"[{i}]", type="int", value=str(i + 1))
                check.check_value(self, child, f"{child}, child {i} of {v.GetName()}")



"""
Test some SBModuleSpec APIs.
"""

import lldb
from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *


class ModuleSpecAPIsTestCase(TestBase):
    def test_object_offset_and_size(self):
        module_spec = lldb.SBModuleSpec()
        self.assertEqual(module_spec.GetObjectOffset(), 0)
        self.assertEqual(module_spec.GetObjectSize(), 0)

        module_spec.SetObjectOffset(4096)
        self.assertEqual(module_spec.GetObjectOffset(), 4096)

        module_spec.SetObjectSize(3600)
        self.assertEqual(module_spec.GetObjectSize(), 3600)

        module_spec.Clear()
        self.assertEqual(module_spec.GetObjectOffset(), 0)
        self.assertEqual(module_spec.GetObjectSize(), 0)



# encoding: utf-8

import lldb


class MyFrameRecognizer(object):
    def get_recognized_arguments(self, frame):
        if frame.name == "foo":
            arg1 = frame.EvaluateExpression("$arg1").signed
            arg2 = frame.EvaluateExpression("$arg2").signed
            val1 = (
                frame.GetThread()
                .GetProcess()
                .GetTarget()
                .CreateValueFromExpression("a", "%d" % arg1)
            )
            val2 = (
                frame.GetThread()
                .GetProcess()
                .GetTarget()
                .CreateValueFromExpression("b", "%d" % arg2)
            )
            return [val1, val2]
        elif frame.name == "bar":
            arg1 = frame.EvaluateExpression("$arg1").signed
            val1 = (
                frame.GetThread()
                .GetProcess()
                .GetTarget()
                .CreateValueFromExpression("a", "(int *)%d" % arg1)
            )
            return [val1]
        return []


class MyOtherFrameRecognizer(object):
    def get_recognized_arguments(self, frame):
        return []


class BazFrameRecognizer(object):
    def should_hide(self, frame):
        return "baz" in frame.name



# encoding: utf-8
"""
Test lldb data formatter subsystem.
"""


import lldb
from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *
from lldbsuite.test import lldbutil

from ObjCDataFormatterTestCase import ObjCDataFormatterTestCase


class ObjCDataFormatterKVO(ObjCDataFormatterTestCase):
    def test_kvo_with_run_command(self):
        """Test the behavior of formatters when KVO is in use."""
        self.build()
        self.target, process, thread, bkpt = lldbutil.run_to_source_breakpoint(
            self, "// Set break point at this line.", lldb.SBFileSpec("main.m", False)
        )

        self.runCmd("settings set target.prefer-dynamic-value no-dynamic-values")

        # The stop reason of the thread should be breakpoint.
        self.expect(
            "thread list",
            STOPPED_DUE_TO_BREAKPOINT,
            substrs=["stopped", "stop reason = breakpoint"],
        )

        # This is the function to remove the custom formats in order to have a
        # clean slate for the next test case.
        def cleanup():
            self.runCmd("type format clear", check=False)
            self.runCmd("type summary clear", check=False)
            self.runCmd("type synth clear", check=False)

        # Execute the cleanup function during test case tear down.
        self.addTearDownHook(cleanup)

        # as long as KVO is implemented by subclassing, this test should succeed
        # we should be able to dynamically figure out that the KVO implementor class
        # is a subclass of Molecule, and use the appropriate summary for it
        self.runCmd("type summary add -s JustAMoleculeHere Molecule")
        self.expect("frame variable molecule", substrs=["JustAMoleculeHere"])
        self.runCmd("next")
        self.expect("thread list", substrs=["stopped", "step over"])
        self.expect("frame variable molecule", substrs=["JustAMoleculeHere"])

        self.runCmd("next")
        # check that NSMutableDictionary's formatter is not confused when
        # dealing with a KVO'd dictionary
        self.expect(
            "frame variable newMutableDictionary",
            substrs=["(NSDictionary *) newMutableDictionary = ", " 21 key/value pairs"],
        )

        lldbutil.run_break_set_by_symbol(self, "-[Molecule setAtoms:]")

        self.runCmd("continue")
        self.expect("frame variable _cmd", substrs=["setAtoms:"])



"""
Test that updating a breakpoint condition correctly invalidates cached state.

This test verifies that when a breakpoint condition is changed, the new condition
is properly evaluated. Previously, due to a bug in StopCondition::SetText where
the hash was computed from a moved-from string, updating conditions could fail
to invalidate cached condition state at breakpoint locations.
"""

import lldb
from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *
from lldbsuite.test import lldbutil


class UpdateBreakpointConditionTestCase(TestBase):
    def setUp(self):
        TestBase.setUp(self)
        self.source = "main.c"

    @add_test_categories(["pyapi"])
    def test_update_condition_python_api(self):
        """Test that updating a breakpoint condition works correctly using Python API."""
        self.build()
        target, process, thread, breakpoint = lldbutil.run_to_source_breakpoint(
            self, "Set breakpoint here", lldb.SBFileSpec(self.source)
        )

        # Set initial condition: x == y.
        breakpoint.SetCondition("x == y")
        self.assertEqual(breakpoint.GetCondition(), "x == y")

        # Need to continue since we're already stopped, but the condition wasn't set initially.
        # First hit should be at foo(5, 5) where x == y.
        process.Continue()
        thread = lldbutil.get_stopped_thread(process, lldb.eStopReasonBreakpoint)
        self.assertTrue(thread.IsValid(), "Should stop at first x == y condition")

        frame = thread.GetFrameAtIndex(0)
        x_val = frame.FindVariable("x")
        y_val = frame.FindVariable("y")
        self.assertEqual(x_val.GetValueAsSigned(), 5, "x should be 5")
        self.assertEqual(y_val.GetValueAsSigned(), 5, "y should be 5")
        self.assertEqual(breakpoint.GetHitCount(), 2)

        # Continue to second hit with x == y.
        process.Continue()
        thread = lldbutil.get_stopped_thread(process, lldb.eStopReasonBreakpoint)
        self.assertTrue(thread.IsValid(), "Should stop at second x == y condition")

        frame = thread.GetFrameAtIndex(0)
        x_val = frame.FindVariable("x")
        y_val = frame.FindVariable("y")
        self.assertEqual(x_val.GetValueAsSigned(), 6, "x should be 6")
        self.assertEqual(y_val.GetValueAsSigned(), 6, "y should be 6")
        self.assertEqual(breakpoint.GetHitCount(), 3)

        # Now update the condition to x > y.
        # This tests the fix for the bug where the hash wasn't updated correctly.
        breakpoint.SetCondition("x > y")
        self.assertEqual(breakpoint.GetCondition(), "x > y")

        # Continue - should now hit at foo(3, 1) where x > y (3 > 1).
        # Without the fix, it would incorrectly hit at foo(7, 7) due to stale condition hash.
        process.Continue()
        thread = lldbutil.get_stopped_thread(process, lldb.eStopReasonBreakpoint)
        self.assertTrue(thread.IsValid(), "Should stop at x > y condition")

        frame = thread.GetFrameAtIndex(0)
        x_val = frame.FindVariable("x")
        y_val = frame.FindVariable("y")
        self.assertEqual(x_val.GetValueAsSigned(), 3, "x should be 3")
        self.assertEqual(y_val.GetValueAsSigned(), 1, "y should be 1")
        self.assertTrue(
            x_val.GetValueAsSigned() > y_val.GetValueAsSigned(),
            "Condition x > y should be true",
        )
        self.assertEqual(breakpoint.GetHitCount(), 4)

    def test_update_condition_command(self):
        """Test that updating a breakpoint condition works correctly using breakpoint modify."""
        self.build()
        target, process, thread, bkpt = lldbutil.run_to_source_breakpoint(
            self, "Set breakpoint here", lldb.SBFileSpec(self.source)
        )

        # Set initial condition: x == y.
        self.runCmd("breakpoint modify -c 'x == y' 1")
        self.expect(
            "breakpoint list",
            substrs=["Condition: x == y"],
        )

        # Continue to first hit at foo(5, 5).
        self.runCmd("continue")
        self.expect("process status", PROCESS_STOPPED, patterns=["Process .* stopped"])
        self.expect(
            "frame variable x y",
            substrs=["x = 5", "y = 5"],
        )

        # Continue to second hit.
        self.runCmd("continue")
        self.expect("process status", PROCESS_STOPPED, patterns=["Process .* stopped"])
        self.expect(
            "frame variable x y",
            substrs=["x = 6", "y = 6"],
        )

        # Update condition to x > y.
        self.runCmd("breakpoint modify -c 'x > y' 1")
        self.expect(
            "breakpoint list",
            substrs=["Condition: x > y"],
        )

        # Continue - should hit at foo(3, 1) where x > y.
        self.runCmd("continue")
        self.expect("process status", PROCESS_STOPPED, patterns=["Process .* stopped"])
        self.expect(
            "frame variable x y",
            substrs=["x = 3", "y = 1"],
        )

        # Verify x > y is actually true.
        self.expect("expr x > y", substrs=["true"])



"""
Test absolute symbols in ELF files to make sure they don't create sections and
to verify that symbol values and size can still be accessed via SBSymbol APIs.
"""

from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *
from lldbsuite.test import lldbutil
import os


class TestAbsoluteSymbol(TestBase):
    @no_debug_info_test
    def test_absolute_symbol(self):
        """
        Load an ELF file that contains two symbols:
        - "absolute" which is a symbol with the section SHN_ABS
        - "main" which is a code symbol in .text

        Index   st_name    st_value           st_size            st_info                             st_other st_shndx Name
        ======= ---------- ------------------ ------------------ ----------------------------------- -------- -------- ===========================
        [    0] 0x00000000 0x0000000000000000 0x0000000000000000 0x00 (STB_LOCAL      STT_NOTYPE   ) 0x00            0
        [    1] 0x00000001 0x0000000000001000 0x0000000000000004 0x12 (STB_GLOBAL     STT_FUNC     ) 0x00            1 main
        [    2] 0x00000006 0xffffffff80000000 0x0000000000000009 0x10 (STB_GLOBAL     STT_NOTYPE   ) 0x00      SHN_ABS absolute

        We used to create sections for symbols whose section ID was SHN_ABS
        and this caused problems as the new sections could interfere with
        with address resolution. Absolute symbols' values are not addresses
        and should not be treated this way.

        New APIs were added to SBSymbol to allow access to the raw integer
        value and size of symbols so symbols whose value was not an address
        could be accessed. Prior to this commit, you could only call:

        SBAddress SBSymbol::GetStartAddress()
        SBAddress SBSymbol::GetEndAddress()

        If the symbol's value was not an address, you couldn't access the
        raw value because the above accessors would return invalid SBAddress
        objects if the value wasn't an address. New APIs were added for this:

        uint64_t SBSymbol::GetValue()
        uint64_t SBSymbol::GetSize();
        """
        src_dir = self.getSourceDir()
        yaml_path = os.path.join(src_dir, "absolute.yaml")
        yaml_base, ext = os.path.splitext(yaml_path)
        obj_path = self.getBuildArtifact("a.out")
        self.yaml2obj(yaml_path, obj_path)

        # Create a target with the object file we just created from YAML
        target = self.dbg.CreateTarget(obj_path)
        self.assertTrue(target, VALID_TARGET)

        module = target.modules[0]

        # Make sure the 'main' symbol is valid and has address values. Also make
        # sure we can access the raw file address and size via the new APIS.
        symbol = module.FindSymbol("main")
        self.assertTrue(symbol.IsValid())
        self.assertTrue(symbol.GetStartAddress().IsValid())
        self.assertTrue(symbol.GetEndAddress().IsValid())
        self.assertEqual(symbol.GetValue(), 0x1000)
        self.assertEqual(symbol.GetSize(), 0x4)

        # Make sure the 'absolute' symbol is valid and has no address values.
        # Also make sure we can access the raw file address and size via the new
        # APIS.
        symbol = module.FindSymbol("absolute")
        self.assertTrue(symbol.IsValid())
        self.assertFalse(symbol.GetStartAddress().IsValid())
        self.assertFalse(symbol.GetEndAddress().IsValid())
        self.assertEqual(symbol.GetValue(), 0xFFFFFFFF80000000)
        self.assertEqual(symbol.GetSize(), 9)

        # Make sure no sections were created for the absolute symbol with a
        # prefix of ".absolute." followed by the symbol name as they interfere
        # with address lookups if they are treated like real sections.
        for section in module.sections:
            self.assertNotEqual(section.GetName(), ".absolute.absolute")



import lldb
from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *
from lldbsuite.test import lldbutil


class TestCase(TestBase):
    NO_DEBUG_INFO_TESTCASE = True

    def run_class_tests(self):
        # Expression not referencing context class.
        self.expect_expr("1 + 1", result_type="int", result_value="2")
        # Referencing context class.
        self.expect_expr("member", result_type="const int", result_value="3")
        # Check the type of context class.
        self.expect_expr("this", result_type="const ContextClass *")

    def test_member_func(self):
        self.build()
        lldbutil.run_to_source_breakpoint(
            self, "// break in function in class.", lldb.SBFileSpec("main.cpp")
        )
        self.run_class_tests()

    def test_templated_member_func(self):
        self.build()
        lldbutil.run_to_source_breakpoint(
            self,
            "// break in templated function in class.",
            lldb.SBFileSpec("main.cpp"),
        )
        self.run_class_tests()

    def run_template_class_tests(self):
        # Expression not referencing context class.
        self.expect_expr("1 + 1", result_type="int", result_value="2")
        # Referencing context class.
        self.expect_expr("member", result_type="const int", result_value="4")
        # Check the type of context class.
        self.expect_expr("this", result_type="const TemplatedContextClass<int> *")

    def test_template_member_func(self):
        self.build()
        lldbutil.run_to_source_breakpoint(
            self,
            "// break in function in templated class.",
            lldb.SBFileSpec("main.cpp"),
        )
        self.run_template_class_tests()

    def test_template_templated_member_func(self):
        self.build()
        lldbutil.run_to_source_breakpoint(
            self,
            "// break in templated function in templated class.",
            lldb.SBFileSpec("main.cpp"),
        )
        self.run_template_class_tests()



"""
Fuzz tests an object after the default construction to make sure it does not crash lldb.
"""

import lldb


def fuzz_obj(obj):
    obj.GetAddress()
    obj.GetLoadAddress()
    obj.SetEnabled(True)
    obj.IsEnabled()
    obj.SetCondition("i >= 10")
    obj.GetCondition()
    obj.SetThreadID(0)
    obj.GetThreadID()
    obj.SetThreadIndex(0)
    obj.GetThreadIndex()
    obj.SetThreadName("worker thread")
    obj.GetThreadName()
    obj.SetQueueName("my queue")
    obj.GetQueueName()
    obj.IsResolved()
    obj.GetDescription(lldb.SBStream(), lldb.eDescriptionLevelVerbose)
    breakpoint = obj.GetBreakpoint()
    # Do fuzz testing on the breakpoint obj, it should not crash lldb.
    import sb_breakpoint

    sb_breakpoint.fuzz_obj(breakpoint)



"""
Test lldb data formatter subsystem.
"""

import lldb
from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *
from lldbsuite.test import lldbutil


class StdFunctionTestCase(TestBase):
    SHARED_BUILD_TESTCASE = False

    # Run frame var for a variable twice. Verify we do not hit the cache
    # the first time but do the second time.
    def run_frame_var_check_cache_use(
        self, variable, result_to_match, skip_find_function=False
    ):
        self.runCmd("log timers reset")
        self.expect(
            "frame variable " + variable, substrs=[variable + " =  " + result_to_match]
        )
        if not skip_find_function:
            self.expect(
                "log timers dump", substrs=["lldb_private::CompileUnit::FindFunction"]
            )

        self.runCmd("log timers reset")
        self.expect(
            "frame variable " + variable, substrs=[variable + " =  " + result_to_match]
        )
        self.expect(
            "log timers dump",
            matching=False,
            substrs=["lldb_private::CompileUnit::FindFunction"],
        )

    def do_test(self):
        """Test that std::function is correctly printed by LLDB"""
        self.runCmd("file " + self.getBuildArtifact("a.out"), CURRENT_EXECUTABLE_SET)

        bkpt = self.target().FindBreakpointByID(
            lldbutil.run_break_set_by_source_regexp(
                self, "Set break point at this line."
            )
        )

        self.runCmd("run", RUN_SUCCEEDED)

        # The stop reason of the thread should be breakpoint.
        self.expect(
            "thread list",
            STOPPED_DUE_TO_BREAKPOINT,
            substrs=["stopped", "stop reason = breakpoint"],
        )

        self.run_frame_var_check_cache_use(
            "foo2_f", "Lambda in File main.cpp at Line 16"
        )

        lldbutil.continue_to_breakpoint(self.process(), bkpt)

        self.run_frame_var_check_cache_use(
            "add_num2_f", "Lambda in File main.cpp at Line 9"
        )

        lldbutil.continue_to_breakpoint(self.process(), bkpt)

        self.run_frame_var_check_cache_use("f2", "Lambda in File main.cpp at Line 26")
        self.run_frame_var_check_cache_use(
            "f3", "Lambda in File main.cpp at Line 30", True
        )
        # TODO reenable this case when std::function formatter supports
        # general callable object case.
        # self.run_frame_var_check_cache_use("f4", "Function in File main.cpp at Line 8")

        # These cases won't hit the cache at all but also don't require
        # an expensive lookup.
        self.expect("frame variable f1", substrs=["f1 =  Function = foo(int, int)"])

        self.expect(
            "frame variable f5", substrs=["f5 =  Function = Bar::add_num(int) const"]
        )

    @add_test_categories(["libc++"])
    def test_libcxx(self):
        self.build(dictionary={"USE_LIBCPP": 1})
        self.do_test()



"""
Test hardware breakpoints for multiple threads.
"""


import lldb
from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *
from lldbsuite.test import lldbutil

from functionalities.breakpoint.hardware_breakpoints.base import *


class HardwareBreakpointMultiThreadTestCase(HardwareBreakpointTestBase):
    @skipTestIfFn(HardwareBreakpointTestBase.hw_breakpoints_unsupported)
    def test_hw_break_set_delete_multi_thread_macos(self):
        self.build()
        self.setTearDownCleanup()
        self.break_multi_thread("delete")

    @skipTestIfFn(HardwareBreakpointTestBase.hw_breakpoints_unsupported)
    def test_hw_break_set_disable_multi_thread_macos(self):
        self.build()
        self.setTearDownCleanup()
        self.break_multi_thread("disable")

    def setUp(self):
        # Call super's setUp().
        TestBase.setUp(self)
        # Our simple source filename.
        self.source = "main.cpp"
        # Find the line number to break inside main().
        self.first_stop = line_number(
            self.source, "Starting thread creation with hardware breakpoint set"
        )

    def break_multi_thread(self, removal_type):
        """Test that lldb hardware breakpoints work for multiple threads."""
        self.runCmd("file " + self.getBuildArtifact("a.out"), CURRENT_EXECUTABLE_SET)

        # Stop in main before creating any threads.
        lldbutil.run_break_set_by_file_and_line(
            self, None, self.first_stop, num_expected_locations=1
        )

        # Run the program.
        self.runCmd("run", RUN_SUCCEEDED)

        # We should be stopped again due to the breakpoint.
        # The stop reason of the thread should be breakpoint.
        self.expect(
            "thread list",
            STOPPED_DUE_TO_BREAKPOINT,
            substrs=["stopped", "stop reason = breakpoint"],
        )

        # Now set a hardware breakpoint in thread function.
        self.expect(
            "breakpoint set -b hw_break_function --hardware",
            substrs=["Breakpoint", "hw_break_function", "address = 0x"],
        )

        # We should stop in hw_break_function function for 4 threads.
        count = 0

        while count < 2:
            self.runCmd("process continue")

            # We should be stopped in hw_break_function
            # The stop reason of the thread should be breakpoint.
            self.expect(
                "thread list",
                STOPPED_DUE_TO_BREAKPOINT,
                substrs=[
                    "hw_break_function",
                    "stop reason = breakpoint",
                ],
            )

            # Continue the loop and test that we are stopped 4 times.
            count += 1

        # Check the breakpoint list.
        self.expect("breakpoint list", substrs=["hw_break_function", "hardware"])
        self.expect(
            "breakpoint list -v",
            substrs=["function = hw_break_function", "hardware = true"],
        )

        if removal_type == "delete":
            self.runCmd("settings set auto-confirm true")

            # Now 'breakpoint delete' should just work fine without confirmation
            # prompt from the command interpreter.
            self.expect("breakpoint delete", startstr="All breakpoints removed")

            # Restore the original setting of auto-confirm.
            self.runCmd("settings clear auto-confirm")

        elif removal_type == "disable":
            self.expect("breakpoint disable", startstr="All breakpoints disabled.")

        # Continue. Program should exit without stopping anywhere.
        self.runCmd("process continue")

        # Process should have stopped and exited with status = 0
        self.expect(
            "process status",
            PROCESS_STOPPED,
            patterns=["Process .* exited with status = 0"],
        )



"""
Test that variables of floating point types are displayed correctly.
"""


import AbstractBase

import lldb
from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *
from lldbsuite.test import lldbutil


class DoubleTypesTestCase(AbstractBase.GenericTester):
    def test_double_type(self):
        """Test that double-type variables are displayed correctly."""
        self.build_and_run("double.cpp", set(["double"]))

    @skipUnlessDarwin
    def test_double_type_from_block(self):
        """Test that double-type variables are displayed correctly from a block."""
        self.build_and_run("double.cpp", set(["double"]), bc=True)



# RUN: %PYTHON %s 2>&1 | FileCheck %s

import gc, os, sys, tempfile
from mlir.ir import *
from mlir.passmanager import *
from mlir.dialects.func import FuncOp
from mlir.dialects.builtin import ModuleOp


# Log everything to stderr and flush so that we have a unified stream to match
# errors/info emitted by MLIR to stderr.
def log(*args):
    print(*args, file=sys.stderr)
    sys.stderr.flush()


def run(f):
    log("\nTEST:", f.__name__)
    f()
    gc.collect()
    assert Context._get_live_count() == 0


# Verify capsule interop.
# CHECK-LABEL: TEST: testCapsule
def testCapsule():
    with Context():
        pm = PassManager()
        pm_capsule = pm._CAPIPtr
        assert '"mlir.passmanager.PassManager._CAPIPtr"' in repr(pm_capsule)
        pm._testing_release()
        pm1 = PassManager._CAPICreate(pm_capsule)
        assert pm1 is not None  # And does not crash.


run(testCapsule)


# CHECK-LABEL: TEST: testConstruct
@run
def testConstruct():
    with Context():
        # CHECK: pm1: 'any()'
        # CHECK: pm2: 'builtin.module()'
        pm1 = PassManager()
        pm2 = PassManager("builtin.module")
        log(f"pm1: '{pm1}'")
        log(f"pm2: '{pm2}'")


# Verify successful round-trip.
# CHECK-LABEL: TEST: testParseSuccess
def testParseSuccess():
    with Context():
        # An unregistered pass should not parse.
        try:
            pm = PassManager.parse(
                "builtin.module(func.func(not-existing-pass{json=false}))"
            )
        except ValueError as e:
            # CHECK: ValueError exception: {{.+}} 'not-existing-pass' does not refer to a registered pass
            log("ValueError exception:", e)
        else:
            log("Exception not produced")

        # A registered pass should parse successfully.
        pm = PassManager.parse("builtin.module(func.func(print-op-stats{json=false}))")
        # CHECK: Roundtrip: builtin.module(func.func(print-op-stats{json=false}))
        log("Roundtrip: ", pm)


run(testParseSuccess)


# Verify successful round-trip.
# CHECK-LABEL: TEST: testParseSpacedPipeline
def testParseSpacedPipeline():
    with Context():
        # A registered pass should parse successfully even if has extras spaces for readability
        pm = PassManager.parse(
            """builtin.module(
        func.func( print-op-stats{ json=false } )
    )"""
        )
        # CHECK: Roundtrip: builtin.module(func.func(print-op-stats{json=false}))
        log("Roundtrip: ", pm)


run(testParseSpacedPipeline)


# Verify failure on unregistered pass.
# CHECK-LABEL: TEST: testParseFail
def testParseFail():
    with Context():
        try:
            pm = PassManager.parse("any(unknown-pass)")
        except ValueError as e:
            #      CHECK: ValueError exception: MLIR Textual PassPipeline Parser:1:1: error:
            # CHECK-SAME: 'unknown-pass' does not refer to a registered pass or pass pipeline
            #      CHECK: unknown-pass
            #      CHECK: ^
            log("ValueError exception:", e)
        else:
            log("Exception not produced")


run(testParseFail)


# Check that adding to a pass manager works
# CHECK-LABEL: TEST: testAdd
@run
def testAdd():
    pm = PassManager("any", Context())
    # CHECK: pm: 'any()'
    log(f"pm: '{pm}'")
    # CHECK: pm: 'any(cse)'
    pm.add("cse")
    log(f"pm: '{pm}'")
    # CHECK: pm: 'any(cse,cse)'
    pm.add("cse")
    log(f"pm: '{pm}'")


# Verify failure on incorrect level of nesting.
# CHECK-LABEL: TEST: testInvalidNesting
def testInvalidNesting():
    with Context():
        try:
            pm = PassManager.parse("func.func(normalize-memrefs)")
        except ValueError as e:
            # CHECK: ValueError exception: Can't add pass 'NormalizeMemRefsPass' restricted to 'builtin.module' on a PassManager intended to run on 'func.func', did you intend to nest?
            log("ValueError exception:", e)
        else:
            log("Exception not produced")


run(testInvalidNesting)


# Verify that a pass manager can execute on IR
# CHECK-LABEL: TEST: testRunPipeline
def testRunPipeline():
    with Context():
        pm = PassManager.parse("any(print-op-stats{json=false})")
        func = FuncOp.parse(r"""func.func @successfulParse() { return }""")
        pm.run(func)


# CHECK: Operations encountered:
# CHECK: func.func      , 1
# CHECK: func.return        , 1
run(testRunPipeline)


# CHECK-LABEL: TEST: testRunPipelineError
@run
def testRunPipelineError():
    with Context() as ctx:
        ctx.allow_unregistered_dialects = True
        op = Operation.parse('"test.op"() : () -> ()')
        pm = PassManager.parse("any(cse)")
        try:
            pm.run(op)
        except MLIRError as e:
            # CHECK: Exception: <
            # CHECK:   Failure while executing pass pipeline:
            # CHECK:   error: "-":1:1: 'test.op' op trying to schedule a pass on an unregistered operation
            # CHECK:    note: "-":1:1: see current operation: "test.op"() : () -> ()
            # CHECK: >
            log(f"Exception: <{e}>")


# CHECK-LABEL: TEST: testPostPassOpInvalidation
@run
def testPostPassOpInvalidation():
    with Context() as ctx:
        module = ModuleOp.parse(
            """
          module {
            arith.constant 10
            func.func @foo() {
              arith.constant 10
              return
            }
          }
        """
        )

        outer_const_op = module.body.operations[0]
        # CHECK: %[[VAL0:.*]] = arith.constant 10 : i64
        log(outer_const_op)

        func_op = module.body.operations[1]
        # CHECK: func.func @[[FOO:.*]]() {
        # CHECK:   %[[VAL1:.*]] = arith.constant 10 : i64
        # CHECK:   return
        # CHECK: }
        log(func_op)

        inner_const_op = func_op.body.blocks[0].operations[0]
        # CHECK: %[[VAL1]] = arith.constant 10 : i64
        log(inner_const_op)

        PassManager.parse("builtin.module(canonicalize)").run(module)
        # CHECK: func.func @foo() {
        # CHECK:   return
        # CHECK: }
        log(func_op)

        # CHECK: func.func @foo() {
        # CHECK:   return
        # CHECK: }
        log(module)

        # CHECK: invalidate_ops=True
        log("invalidate_ops=True")

        module = ModuleOp.parse(
            """
          module {
            arith.constant 10
            func.func @foo() {
              arith.constant 10
              return
            }
          }
        """
        )

        PassManager.parse("builtin.module(canonicalize)").run(module)

        func_op._set_invalid()
        try:
            log(func_op)
        except RuntimeError as e:
            # CHECK: the operation has been invalidated
            log(e)

        outer_const_op._set_invalid()
        try:
            log(outer_const_op)
        except RuntimeError as e:
            # CHECK: the operation has been invalidated
            log(e)

        inner_const_op._set_invalid()
        try:
            log(inner_const_op)
        except RuntimeError as e:
            # CHECK: the operation has been invalidated
            log(e)

        # CHECK: func.func @foo() {
        # CHECK:   return
        # CHECK: }
        log(module)


# CHECK-LABEL: TEST: testPrintIrAfterAll
@run
def testPrintIrAfterAll():
    with Context() as ctx:
        module = ModuleOp.parse(
            """
          module {
            func.func @main() {
              %0 = arith.constant 10
              return
            }
          }
        """
        )
        pm = PassManager.parse("builtin.module(canonicalize)")
        ctx.enable_multithreading(False)
        pm.enable_ir_printing()
        # CHECK: // -----// IR Dump After CanonicalizerPass (canonicalize) //----- //
        # CHECK: module {
        # CHECK:   func.func @main() {
        # CHECK:     return
        # CHECK:   }
        # CHECK: }
        pm.run(module)


# CHECK-LABEL: TEST: testPrintIrBeforeAndAfterAll
@run
def testPrintIrBeforeAndAfterAll():
    with Context() as ctx:
        module = ModuleOp.parse(
            """
          module {
            func.func @main() {
              %0 = arith.constant 10
              return
            }
          }
        """
        )
        pm = PassManager.parse("builtin.module(canonicalize)")
        ctx.enable_multithreading(False)
        pm.enable_ir_printing(print_before_all=True, print_after_all=True)
        # CHECK: // -----// IR Dump Before CanonicalizerPass (canonicalize) //----- //
        # CHECK: module {
        # CHECK:   func.func @main() {
        # CHECK:     %[[C10:.*]] = arith.constant 10 : i64
        # CHECK:     return
        # CHECK:   }
        # CHECK: }
        # CHECK: // -----// IR Dump After CanonicalizerPass (canonicalize) //----- //
        # CHECK: module {
        # CHECK:   func.func @main() {
        # CHECK:     return
        # CHECK:   }
        # CHECK: }
        pm.run(module)


# CHECK-LABEL: TEST: testPrintIrLargeLimitElements
@run
def testPrintIrLargeLimitElements():
    with Context() as ctx:
        module = ModuleOp.parse(
            """
          module {
            func.func @main() -> tensor<3xi64> {
              %0 = arith.constant dense<[1, 2, 3]> : tensor<3xi64>
              return %0 : tensor<3xi64>
            }
          }
        """
        )
        pm = PassManager.parse("builtin.module(canonicalize)")
        ctx.enable_multithreading(False)
        pm.enable_ir_printing(large_elements_limit=2)
        # CHECK:     %[[CST:.*]] = arith.constant dense_resource<__elided__> : tensor<3xi64>
        pm.run(module)


# CHECK-LABEL: TEST: testPrintIrLargeResourceLimit
@run
def testPrintIrLargeResourceLimit():
    with Context() as ctx:
        module = ModuleOp.parse(
            """
          module {
            func.func @main() -> tensor<3xi64> {
              %0 = arith.constant dense_resource<blob1> : tensor<3xi64>
              return %0 : tensor<3xi64>
            }
          }
          {-#
            dialect_resources: {
              builtin: {
                blob1: "0x010000000000000002000000000000000300000000000000"
              }
            }
          #-}
        """
        )
        pm = PassManager.parse("builtin.module(canonicalize)")
        ctx.enable_multithreading(False)
        pm.enable_ir_printing(large_resource_limit=4)
        # CHECK-NOT: blob1: "0x01
        pm.run(module)


# CHECK-LABEL: TEST: testPrintIrLargeResourceLimitVsElementsLimit
@run
def testPrintIrLargeResourceLimitVsElementsLimit():
    """Test that large_elements_limit does not affect the printing of resources."""
    with Context() as ctx:
        module = ModuleOp.parse(
            """
          module {
            func.func @main() -> tensor<3xi64> {
              %0 = arith.constant dense_resource<blob1> : tensor<3xi64>
              return %0 : tensor<3xi64>
            }
          }
          {-#
            dialect_resources: {
              builtin: {
                blob1: "0x010000000000000002000000000000000300000000000000"
              }
            }
          #-}
        """
        )
        pm = PassManager.parse("builtin.module(canonicalize)")
        ctx.enable_multithreading(False)
        pm.enable_ir_printing(large_elements_limit=1)
        # CHECK-NOT: blob1: "0x01
        pm.run(module)


# CHECK-LABEL: TEST: testPrintIrTree
@run
def testPrintIrTree():
    with Context() as ctx:
        module = ModuleOp.parse(
            """
          module {
            func.func @main() {
              %0 = arith.constant 10
              return
            }
          }
        """
        )
        pm = PassManager.parse("builtin.module(canonicalize)")
        ctx.enable_multithreading(False)
        pm.enable_ir_printing()
        # CHECK-LABEL: // Tree printing begin
        # CHECK: \-- builtin_module_no-symbol-name
        # CHECK:     \-- 0_canonicalize.mlir
        # CHECK-LABEL: // Tree printing end
        pm.run(module)
        log("// Tree printing begin")
        with tempfile.TemporaryDirectory() as temp_dir:
            pm.enable_ir_printing(tree_printing_dir_path=temp_dir)
            pm.run(module)

            def print_file_tree(directory, prefix=""):
                entries = sorted(os.listdir(directory))
                for i, entry in enumerate(entries):
                    path = os.path.join(directory, entry)
                    connector = "\-- " if i == len(entries) - 1 else "|-- "
                    log(f"{prefix}{connector}{entry}")
                    if os.path.isdir(path):
                        print_file_tree(
                            path, prefix + ("    " if i == len(entries) - 1 else "│   ")
                        )

            print_file_tree(temp_dir)
        log("// Tree printing end")


# CHECK-LABEL: TEST: testEnableStatistics
@run
def testEnableStatistics():
    with Context() as ctx:
        module = ModuleOp.parse(
            """
          module {
            func.func @main() {
              %0 = arith.constant 10
              return
            }
          }
        """
        )
        pm = PassManager.parse("builtin.module(canonicalize)")
        pm.enable_statistics()
        # CHECK: Pass statistics report
        pm.run(module)
